#!/usr/bin/env perl
#
# 02-Nov-2009, KAB - this script applies changes to C++
# classes generated by the XSD C++-to-XML compiler to provide the
# user interface that we want.

use Getopt::Std;
use File::Basename;
use strict;

# parse any command-line switches
use vars qw($opt_h $opt_p $opt_r);
my $optionSuccess = getopts("hp:r:");

# if the option parsing failed or the user requested help or there are
# command-line arguments other than switches, print out a usage message
if (! $optionSuccess || $opt_h || @ARGV < 1)
{
    print "Usage:  tailorXsdGeneratedClassesForNova.pl [-h] [-r regFile] [-p pkg] <file1> [file2 [file3 ..]]\n";
    print " where  -h : Help (this message)\n";
    print "        -r : Registration file - modified to support registration of the generated classes\n";
    print "        -p : package name to prefix included headers in registration file\n";
    exit;
}

my ($xsd_version_major) = ($ENV{CSTXSD_VERSION} =~ m&^v(\d+)&);

# Loop over each file
my %classNames;
my @fileList = @ARGV;
foreach my $fileName (@fileList) {

    my $inputFile = $fileName;
    my $outputFile = "${inputFile}.modified";
    my $threeLineWindow = "";
    my $sixLineWindow = "";

    # initial scan through the file to pick up needed relationships
    my %anonTypeElements;
    my %namedTypeElements;
    my %destructorList;
    open FILEIN, $inputFile or die "Unable to open input file $inputFile\n.";
    $threeLineWindow = "";
    $sixLineWindow = "";
    while (my $line = <FILEIN>) {
        chomp $line;

        # update the multi-line "windows" with this new line
        if ($threeLineWindow =~ m/(.*)\n(.*)\n(.*)/) {
            $threeLineWindow = $2 . "\n" . $3;
        }
        $threeLineWindow .= "\n" . $line;
        if ($sixLineWindow =~ m/(.*)\n(.*)\n(.*)\n(.*)\n(.*)\n(.*)/) {
            $sixLineWindow = $2 . "\n" . $3 . "\n" . $4 . "\n" . $5 . "\n" . $6;
        }
        $sixLineWindow .= "\n" . $line;

        # identify elements that have been defined with anonymous types
        # (we use the simplest global function that serializes elements
        # to existing DOMDocuments to do this)
        if ($sixLineWindow =~
            m/\s*void\s+(\S+)_\s*\(::xercesc::DOMDocument&\s*\w*,\s*const\s+::(\S*)\1&\s*\w*,\s*::xml_schema::flags/s) {
            my $globalFunctionName = $1 . "_";
            my $fullyQualifiedClassName = $2 . $1;
            my $className = $fullyQualifiedClassName;
            $className =~ s/.*:://;
            $anonTypeElements{$globalFunctionName} = $fullyQualifiedClassName;
            $classNames{$fullyQualifiedClassName} = $className;
        }

        # identify elements that have been defined with named types
        elsif ($sixLineWindow =~
            m/\s*void\s+(\S+)\s*\(::xercesc::DOMDocument&\s*\w*,\s*const\s+::(\S+)\&\s*\w*,\s*::xml_schema::flags/s) {
            my $globalFunctionName = $1;
            my $fullyQualifiedClassName = $2;
            my $className = $fullyQualifiedClassName;
            $className =~ s/.*:://;
            $namedTypeElements{$globalFunctionName} = $fullyQualifiedClassName;
            $classNames{$fullyQualifiedClassName} = $className;
        }

        # identify classes (by taking note of the presence of a destructor)
        if ($line =~ m/\s*\~(\S+)\s*\(/) {
            $destructorList{$1} = 1;
        }
    }
    close FILEIN;

    #foreach my $key (sort(keys(%anonTypeElements))) {
    #    my $val = $anonTypeElements{$key};
    #    print STDOUT "file = $inputFile, anon element = $key, type = $val\n";
    #}
    #foreach my $key (sort(keys(%namedTypeElements))) {
    #    my $val = $namedTypeElements{$key};
    #    print STDOUT "file = $inputFile, named element = $key, type = $val\n";
    #}

    # now, on to modifying the file...
    open FILEIN, $inputFile or die "Unable to open input file $inputFile\n.";
    open FILEOUT, ">$outputFile" or die "Unable to open output file $outputFile\n";

    my $topLevelClass = "";
    my $currentGeneratedClass = "none";
    my %nonArrayDataMembers;
    my %arrayDataMembers;
    my $copyTraitsHaveBeenGenerated = 0;
    $threeLineWindow = "";
    $sixLineWindow = "";
    while (my $line = <FILEIN>) {
        chomp $line;

        # update the multi-line "windows" with this new line
        if ($threeLineWindow =~ m/(.*)\n(.*)\n(.*)/) {
            $threeLineWindow = $2 . "\n" . $3;
        }
        $threeLineWindow .= "\n" . $line;
        if ($sixLineWindow =~ m/(.*)\n(.*)\n(.*)\n(.*)\n(.*)\n(.*)/) {
            $sixLineWindow = $2 . "\n" . $3 . "\n" . $4 . "\n" . $5 . "\n" . $6;
        }
        $sixLineWindow .= "\n" . $line;
        #print STDOUT "========================================\n";
        #print STDOUT "$sixLineWindow\n";

        # header file modifications
        if ($inputFile =~ /h$/) {

            # add #include for XMLSerializable and XMLDeserializationRegistry...
            if ($threeLineWindow =~ m/\#include\s+\<\S+\>\s+namespace\s+xml_schema\s*\{?\s*$/s) {
                $line = "#include <NovaDAQUtilities/XMLDeserializationRegistry.h>\n\n" . $line;

                # 22-Feb-2011, KAB - additional includes
                $line = "#include <iostream>\n#include <fstream>\n#include <stdint.h>\n" . $line;
            }

            # track whether we are working on a top-level class
            # also, add the XMLSerializable base class to top-level classes
            # (for reference, "top-level" classes are ones that actually get
            # serialized into messages - they are defined as "elements" in the xsd)
            if ($line =~ m/(\s*)class\s+(\w+)\s*:\s*public\s+::xml_schema::type(\s*\{?\s*)/) {
                my $indentation = $1;
                my $thisClassName = $2;
                my $trailingText = $3;
                $topLevelClass = "";
                if (! $topLevelClass) {
                    foreach my $fullyQualifiedClassName (sort(values(%anonTypeElements))) {
                        my $testClassName = $classNames{$fullyQualifiedClassName};
                        if ("$testClassName" eq "$thisClassName") {
                            $topLevelClass = $testClassName;
                            last;
                        }
                    }
                }
                if (! $topLevelClass) {
                    foreach my $fullyQualifiedClassName (sort(values(%namedTypeElements))) {
                        my $testClassName = $classNames{$fullyQualifiedClassName};
                        if ("$testClassName" eq "$thisClassName") {
                            $topLevelClass = $testClassName;
                            last;
                        }
                    }
                }
                if ($topLevelClass) {
                    $line .= "\n" . $indentation . "      " .
                        ", public ::novadaq::XMLSerializable" .
                        $trailingText;
                }
            }

            # add needed methods to top-level classes
            if ($line =~ m/^\s*$/ &&
                $threeLineWindow =~ m/(\s*)virtual\s+\~(\w+)\s*\(\s*\)\s*;/s) {
                $line .= "\n" . $1 . "void copyNeededFields (const " . $2 . "& x);\n";
                if ($topLevelClass) {
                    $line .= "\n" . $1 . "::std::string serialize () const;\n\n" .
                        $1 . "void deserialize (const ::std::string&);\n\n" .
                        $1 . "virtual ::std::string getClassName () const;\n\n" .
                        $1 . "virtual ::boost::shared_ptr< ::novadaq::" .
                        "XMLSerializable > clone () const;\n\n" .
                        $1 . "static void registerForDeserialization ();\n";
                }
            }

            # add "getXYZ" methods (to retrieve individual attributes)
            if ($line =~ m/^\s*$/ &&
                $threeLineWindow =~ m/(\s*)(\w+)_type\s*\&\s*\2\s*\(\s*\);\s+/s) {
                $line .= "\n" . $1 . "const " . $2 . "_type&\n" .
                    $1 . "get" . ucfirst($2) . " () const;\n\n" .
                    $1 . "bool\n" .
                    $1 . "has" . ucfirst($2) . " () const;\n";
            }

            # add "getXYZ" methods (to retrieve elements from an attribute which is a list)
            if ($line =~ m/^\s*$/ &&
                $threeLineWindow =~ m/(\s*)(\w+)_sequence\s*\&\s*\2\s*\(\s*\);\s+/s) {
                $line .= "\n" . $1 . "::xml_schema::integer\n" .
                    $1 . "get" . ucfirst($2) . "Count () const;\n";

                $line .= "\n" . $1 . "const " . $2 . "_type&\n" .
                    $1 . "get" . ucfirst($2) . " (int index) const;\n";

                $line .= "\n" . $1 . $2 . "_type&\n" .
                    $1 . "get" . ucfirst($2) . "Reference (int index);\n";

                $line .= "\n" . $1 . "bool\n" .
                    $1 . "has" . ucfirst($2) . " () const;\n";

                $line .= "\n" . $1 . "void\n" .
                    $1 . "removeAll" . ucfirst($2) . " ();\n";
            }

            # add "getXYZ" methods (to retrieve elements from an attribute which is optional)
            if ($line =~ m/^\s*$/ &&
                $threeLineWindow =~ m/(\s*)(\w+)_optional\s*\&\s*\2\s*\(\s*\);\s+/s) {
                $line .= "\n" . $1 . "const " . $2 . "_type&\n" .
                    $1 . "get" . ucfirst($2) . " () const;\n";

                $line .= "\n" . $1 . "bool\n" .
                    $1 . "has" . ucfirst($2) . " () const;\n";
            }

            # add "setXYZ" methods (to set individual attributes)
            if ($line =~ m/^\s*$/ &&
                $threeLineWindow =~ m/(\s*)void\s+(\w+)\s*\(const\s+\2_type\s*\&(\s*\w*\s*)\);\s+/s) {
                $line .= "\n" . $1 . "void\n" .
                    $1 . "set" . ucfirst($2) . " (const " . $2 . "_type&" . $3 . ");\n";
            }

            # add "addXYZ" methods (to add elements to an attribute which is a list)
            if ($line =~ m/^\s*$/ &&
                $threeLineWindow =~ m/(\s*)void\s+(\w+)\s*\(\s*const\s*\2_sequence\s*\&(\s*\w*\s*)\);\s+/s) {
                $line .= "\n" . $1 . "void\n" .
                    $1 . "add" . ucfirst($2) . " (const " . $2 . "_type& new" . ucfirst($2) . ");\n"
            }
        }

        # source file modifications
        if ($inputFile =~ /cpp$/) {

            # add the copyTraits templates
            if ($copyTraitsHaveBeenGenerated == 0 && ! ($line =~ m/namespace/) &&
                ! ($line =~ m/\{/) && $threeLineWindow =~ m/^(\s*)namespace/s) {

                $copyTraitsHaveBeenGenerated = 1;
                my $spaces = $1;

                my $newText = $spaces . "  template <class X>" .
                    "\n" . $spaces . "  struct copyTraits" .
                    "\n" . $spaces . "  {" .
                    "\n" . $spaces . "    static void copyNeededFields(X const& source, " .
                    "X& target) {}" .
                    "\n" . $spaces . "    template <class Y>" .
                    "\n" . $spaces . "    static void copyNeededFields(X const& source, " .
                    "Y& target) {}" .
                    "\n" . $spaces . "  };";
                for my $className (sort(keys(%destructorList))) {
                    $newText .= "\n\n" . $spaces . "  template <>" .
                        "\n" . $spaces . "  struct copyTraits<" . $className . ">" .
                        "\n" . $spaces . "  {" .
                        "\n" . $spaces . "    static void copyNeededFields(" . $className .
                        " const& source, " . $className . "& target) {" .
                        "\n" . $spaces . "      target.copyNeededFields(source);" .
                        "\n" . $spaces . "    }" .
                        "\n" . $spaces . "    template <class Y>" .
                        "\n" . $spaces . "    static void copyNeededFields(" . $className .
                        " const& source, Y& target) {" .
                        "\n" . $spaces . "      target->copyNeededFields(source);" .
                        "\n" . $spaces . "    }" .
                        "\n" . $spaces . "  };";
                }

                $line = $newText . "\n\n" . $line;
            }

            # add "getXYZ" methods (to retrieve individual attributes)
            if ($line =~ m/^\s*$/ &&
                $sixLineWindow =~ m/^(\s*)(\w+)::(\w+)_type\s*\&\s*\2::\s*\3\s*\(\s*\)(.+)/s) {
                my $spaces = $1;
                my $className = $2;
                my $attrName = $3;
                my $methodBody = $4;
                $line .= "\n" . $spaces . "const " . $className . "::" . $attrName . "_type& " . $className . "::\n" .
                    $spaces . "get" . ucfirst($attrName) . " () const" . $methodBody;

                $methodBody =~ s/get/present/;
                $line .= "\n" . $spaces . "bool " . $className . "::\n" .
                    $spaces . "has" . ucfirst($attrName) . " () const" . $methodBody;
            }

            # add "getXYZ" methods (to retrieve elements from an attribute which is a list)
            if ($line =~ m/^\s*$/ &&
                $sixLineWindow =~ m/^(\s*)(\w+)::(\w+)_sequence\s*\&\s*\2::\s*\3\s*\(\s*\)(.+)/s) {
                $line .= "\n" . $1 . "::xml_schema::integer " . $2 . "::\n" .
                    $1 . "get" . ucfirst($3) . "Count () const\n" .
                    $1 . "{\n" .
                    $1 . "  return this->" . $3 . "_.size ();\n" .
                    $1 . "}\n";

                $line .= "\n" . $1 . "const " . $2 . "::" . $3 . "_type& " . $2 . "::\n" .
                    $1 . "get" . ucfirst($3) . " (int index) const\n" .
                    $1 . "{\n" .
                    $1 . "  return this->" . $3 . "_[index];\n" .
                    $1 . "}\n";

                $line .= "\n" . $1 . $2 . "::" . $3 . "_type& " . $2 . "::\n" .
                    $1 . "get" . ucfirst($3) . "Reference (int index)\n" .
                    $1 . "{\n" .
                    $1 . "  return this->" . $3 . "_[index];\n" .
                    $1 . "}\n";

                $line .= "\n" . $1 . "bool " . $2 . "::\n" .
                    $1 . "has" . ucfirst($3) . " () const\n" .
                    $1 . "{\n" .
                    $1 . "  return ! (this->" . $3 . "_.empty ());\n" .
                    $1 . "}\n";

                $line .= "\n" . $1 . "void " . $2 . "::\n" .
                    $1 . "removeAll" . ucfirst($3) . " ()\n" .
                    $1 . "{\n" .
                    $1 . "  this->" . $3 . "_.clear ();\n" .
                    $1 . "}\n";
            }

            # add "getXYZ" methods (to retrieve elements from an attribute which is optional)
            if ($line =~ m/^\s*$/ &&
                $sixLineWindow =~ m/^(\s*)(\w+)::(\w+)_optional\s*\&\s*\2::\s*\3\s*\(\s*\)(.+)/s) {
                $line .= "\n" . $1 . "const " . $2 . "::" . $3 . "_type& " . $2 . "::\n" .
                    $1 . "get" . ucfirst($3) . " () const\n" .
                    $1 . "{\n" .
                    $1 . "  return this->" . $3 . "_.get ();\n" .
                    $1 . "}\n";
                $line .= "\n" . $1 . "bool " . $2 . "::\n" .
                    $1 . "has" . ucfirst($3) . " () const\n" .
                    $1 . "{\n" .
                    $1 . "  return this->" . $3 . "_.present ();\n" .
                    $1 . "}\n";
            }

            # add "setXYZ" methods (to set individual and optional attributes)
            if ($sixLineWindow =~ m/(\s*)void (\w+)::\s+(\w+)\s*\(const\s+\3_type\s*\&(.+)/s) {
                my $newText .= "\n" . $1 . "void " . $2 . "::\n" .
                    $1 . "set" . ucfirst($3) . " (const " . $3 . "_type&" . $4;
                if ($line =~ m/^\s*$/) {
                    $line .= $newText;
                }
                elsif ($sixLineWindow =~ m/\s*\}\s*\}\s*$/s) {
                    $line = $newText;
                }
            }

            # add "addXYZ" methods (to add elements to an attribute which is a list)
            if ($sixLineWindow =~ m/(\s*)void\s+(\w+)::\s*(\w+)\s*\(\s*const\s*\3_sequence\s*\&(.*)/s) {
                my $newText .= "\n" . $1 . "void " . $2 . "::\n" .
                    $1 . "add" . ucfirst($3) . " (const " . $2 . "::" . $3 . "_type& new" . ucfirst($3) . ")\n" .
                    $1 . "{\n" .
                    $1 . "  this->" . $3 . "_.push_back (new" . ucfirst($3) . ");\n" .
                    $1 . "}\n";
                if ($line =~ m/^\s*$/) {
                    $line .= $newText;
                }
                elsif ($sixLineWindow =~ m/\s*\}\s*\}\s*$/s) {
                    $line = $newText . $line;
                }
            }

            # add #includes needed for deserialization
            if ($line =~ m/include.+parsing-source.hxx/s) {
                $line .= "\n#include <xsd/cxx/xml/sax/std-input-source.hxx>" .
                    "\n#include <xsd/cxx/tree/error-handler.hxx>";
            }

            # add needed methods to top-level classes
            if ($sixLineWindow =~ m/([ \t]*)(\w+)\s*::\s*\~\2\s*\(\s*\)\s*\{\s*\}[ \t]*\z/s) {
                my $indentation = $1;
                my $thisClassName = $2;
                $topLevelClass = "";
                my $fullClassName;
                my $helperName;
                if (! $topLevelClass) {
                    foreach my $functionName (sort(keys(%anonTypeElements))) {
                        my $fullyQualifiedClassName = $anonTypeElements{$functionName};
                        my $testClassName = $classNames{$fullyQualifiedClassName};
                        if ("$testClassName" eq "$thisClassName") {
                            $topLevelClass = $testClassName;
                            $fullClassName = $fullyQualifiedClassName;
                            $helperName = $functionName;
                            last;
                        }
                    }
                }
                if (! $topLevelClass) {
                    foreach my $functionName (sort(keys(%namedTypeElements))) {
                        my $fullyQualifiedClassName = $namedTypeElements{$functionName};
                        my $testClassName = $classNames{$fullyQualifiedClassName};
                        if ("$testClassName" eq "$thisClassName") {
                            $topLevelClass = $testClassName;
                            $fullClassName = $fullyQualifiedClassName;
                            $helperName = $functionName;
                            last;
                        }
                    }
                }
                if ($topLevelClass) {
                    $line .= "\n\n" . $indentation . "::std::string " . $thisClassName . "::\n" .
                        $indentation . "serialize () const\n" .
                        $indentation . "{\n" .
                        $indentation . "  // To avoid problems with multi-threaded access to\n" .
                        $indentation . "  // CST XSD ser/deserialization code and underlying\n" .
                        $indentation . "  // Xerces-C++ code, we prevent concurrent access\n" .
                        $indentation . "  // to any serialization code by any class that\n" .
                        $indentation . "  // inherits from XMLSerializable via the use of a\n" .
                        $indentation . "  // static mutex in XMLSerializable.\n" .
                        $indentation . "  boost::mutex::scoped_lock lock(_serializationMutex);\n" .
                        "\n" .
                        $indentation . "  ::std::ostringstream oss;\n" .
                        $indentation . "  " . $helperName . " (oss, *this);\n" .
                        $indentation . "  return oss.str();\n" .
                        $indentation . "}\n";

                    $line .= "\n" . $indentation . "void " . $thisClassName . "::\n" .
                        $indentation . "deserialize (const ::std::string& serializedData)\n" .
                        $indentation . "{\n" .
                        $indentation . "  // To avoid problems with multi-threaded access to\n" .
                        $indentation . "  // CST XSD ser/deserialization code and underlying\n" .
                        $indentation . "  // Xerces-C++ code, we prevent concurrent access\n" .
                        $indentation . "  // to any deserialization code by any class that\n" .
                        $indentation . "  // inherits from XMLSerializable via the use of a\n" .
                        $indentation . "  // static mutex in XMLSerializable.\n" .
                        $indentation . "  boost::mutex::scoped_lock lock(_serializationMutex);\n" .
                        "\n" .
                        $indentation . "  // initialization for XSD\n" .
                        $indentation . "  ::xml_schema::flags flags = ::xml_schema::flags::dont_validate;\n" .
                        $indentation . "  ::xsd::cxx::xml::auto_initializer autoInit\n" .
                        $indentation . "        ((flags & ::xml_schema::flags::dont_initialize) == 0,\n" .
                        $indentation . "         (flags & ::xml_schema::flags::keep_dom) == 0);\n" .
                        $indentation . "  ::xsd::cxx::tree::error_handler< char > errorHandler;\n" .
                        $indentation . "  ::xml_schema::properties props;\n" .
                        "\n" .
                        $indentation . "  // wrap the string in an input stream\n" .
                        $indentation . "  ::std::istringstream iss (serializedData);\n" .
                        $indentation . "  ::xsd::cxx::xml::sax::std_input_source isrc (iss);\n" .
                        "\n" .
                        $indentation . "  // create a DOM from the input stream\n" .
                        $indentation . "  ::xml_schema::dom::auto_ptr< ::xercesc::DOMDocument > doc (\n" .
                        $indentation . "      ::xsd::cxx::xml::dom::parse< char > (isrc, errorHandler, props, flags));\n" .
                        $indentation . "  errorHandler.throw_if_failed< ::xsd::cxx::tree::parsing< char > > ();\n" .
                        "\n" .
                        $indentation . "  // do the deserialization\n";
                    if ($xsd_version_major > 3) {
                        $line .= $indentation . "  ::xsd::cxx::xml::dom::parser< char > parser (*(doc->getDocumentElement()), true, false, true);\n";
                    } else {
                        $line .= $indentation . "  ::xsd::cxx::xml::dom::parser< char > parser (*(doc->getDocumentElement()), true, true);\n";
                    }
                    $line .= $indentation . "  this->parse (parser, flags);\n" .
                        $indentation . "}\n";

                    $line .= "\n" . $indentation . "::std::string " . $thisClassName . "::\n" .
                        $indentation . "getClassName () const\n" .
                        $indentation . "{\n" .
                        $indentation . "  return \"" . $fullClassName . "\";\n" .
                        $indentation . "}\n";

                    $line .= "\n" . $indentation . "::boost::shared_ptr< " .
                        "::novadaq::XMLSerializable > " . $thisClassName . "::\n" .
                        $indentation . "clone () const\n" .
                        $indentation . "{\n" .
                        $indentation . "  " . $thisClassName . " *copy = this->_clone ();\n" .
                        $indentation . "  boost::shared_ptr<novadaq::XMLSerializable> copyPtr (copy);\n" .
                        $indentation . "  return copyPtr;\n" .
                        $indentation . "}\n";

                    $line .= "\n" . $indentation . "void " . $thisClassName . "::\n" .
                        $indentation . "registerForDeserialization ()\n" .
                        $indentation . "{\n" .
                        $indentation . "  boost::shared_ptr<novadaq::XMLSerializable> msgPtr;\n" .
                        $indentation . "  novadaq::XMLDeserializationRegistry&\n" .
                        $indentation . "    deserRegistry = novadaq::XMLDeserializationRegistry::getInstance();\n" .
                        $indentation . "  msgPtr.reset(new " . $thisClassName . "());\n" .
                        $indentation . "  deserRegistry.registerClass(msgPtr->getClassName(), msgPtr);\n" .
                        $indentation . "}";
                }
            }

            # track what type of operator<< method we are in, if any, to provide a
            # nice warning message below
            if ($threeLineWindow =~
                m/\s*void\s+operator<<\s*\(::xercesc::DOM.+const\s*(\w+)\&/s) {
                $currentGeneratedClass = $1;
            }

            # add warning messages about missing fields in the serialization code
            if ($sixLineWindow =~ m/::xercesc::DOM.*&\s*\w+.+::xsd::cxx::xml::dom::create_/s &&
                $line =~ m/(\s*)\w+\s*<<\s*i\.(\w+)\s*\(\s*\)/) {
                # skip fields that appear to be constants by virtue of the
                # fact that their names are in all capital letters
                if (uc($2) ne $2) {
                    $line = $1 . "if (! i.has" . ucfirst($2) . " ()) {\n" .
                        $1 . "  std::cerr << \"WARNING: attempting to serialize a " .
                        $currentGeneratedClass. " with an empty " . $2 .
                        " field!\" << std::endl;\n" .
                        $1 . "}\n" . $line;
                }
            }

            # track what type of parse method we are in, if any, to provide a
            # nice warning message below
            # --> and track the data members that are referenced in the parse method
            # so that we can correctly build the copyNeededFields method below
            if ($threeLineWindow =~
                m/\s*void\s+(\S+)\s*::\s*parse.*::xml_schema::flags f\)/s) {
                $currentGeneratedClass = $1;
                %nonArrayDataMembers = ();
                %arrayDataMembers = ();
            }
            if ($line =~ m/\s*if\s*\(\s*!\s*this\s*->\s*(\S+)\s*\)/) {
                $nonArrayDataMembers{$1} = 1;
            }
            if ($line =~ m/\s*if\s*\(\s*!\s*(\S+)\.present/) {
                # skip fields that appear to be constants by virtue of the
                # fact that their names are in all capital letters
                if (uc($1) ne $1) {
                    $nonArrayDataMembers{$1} = 1;
                }
            }
            if ($line =~ m/\s*this\s*\-\>\s*(\S+).push_back\s*\(/) {
                $arrayDataMembers{$1} = 1;
            }

            # add warning messages about missing fields in the parsing code
            if ($threeLineWindow =~
                m/(\s*)if\s*\(\s*\!\s*(\S+)\_\s*\.\s*present\s*\(\s*\)\s*\)\s*\{\s*throw\s*::xsd::cxx::tree::expected_(\w+)\<\s*char\s*\>\s*\(/) {
                $line = $1 . "  if ((f & ::novadaq::XMLSerializable::" .
                    "allow_missing_elements) == 0)\n" . $1 . "  {\n" .
                    $1 . "    std::cerr << \"ERROR: expected " . $3 . " (" .
                    $currentGeneratedClass . "::" . $2 .
                    ") not found!\" << std::endl;\n  " . $line;
                my $tmpLine;
                if ($tmpLine = <FILEIN>) {
                    chomp $tmpLine;
                    $line .= "\n  " . $tmpLine;
                }
                if ($tmpLine = <FILEIN>) {
                    chomp $tmpLine;
                    $line .= "\n  " . $tmpLine;
                }
                $line .= "\n" . $1 . "  }";
            }

            # add the copyNeededFields method
            if ($line =~ m/\}/ && $threeLineWindow =~
                m/(\s*)\~${currentGeneratedClass}\s*\(\s*\)\s*\{/s) {
                my $spaces = $1;
                $line .= "\n\n" . $spaces . "void " . ${currentGeneratedClass} . "::" .
                    "\n" . $spaces . "copyNeededFields (const " .
                    ${currentGeneratedClass} . "& x)" .
                    "\n" . $spaces . "{";
                for my $fieldName (sort(keys(%nonArrayDataMembers))) {
                    $fieldName =~ s/\_$//;
                    $line .= "\n" . $spaces . "  if (x.has" . ucfirst($fieldName) . "()) {" .
                        "\n" . $spaces . "    if (! this->has" . ucfirst($fieldName) . "()) {" .
                        "\n" . $spaces . "      this->set" . ucfirst($fieldName) .
                        "(x.get" . ucfirst($fieldName) . "());" .
                        "\n" . $spaces . "    }" .
                        "\n" . $spaces . "    else {" .
                        "\n" . $spaces . "      copyTraits<" . $fieldName .
                        "_type>::copyNeededFields(x.get" .
                        ucfirst($fieldName) . "(), this->" . $fieldName . "());" .
                        "\n" . $spaces . "    }" .
                        "\n" . $spaces . "  }";
                }
                for my $fieldName (sort(keys(%arrayDataMembers))) {
                    $fieldName =~ s/\_$//;
                    $line .= "\n" . $spaces . "  if (x.has" . ucfirst($fieldName) . "() && x.get" .
                        ucfirst($fieldName) . "Count() > 0) {" .
                        "\n" . $spaces . "    if (! this->has" . ucfirst($fieldName) .
                        "() || this->get" . ucfirst($fieldName) . "Count() == 0) {" .
                        "\n" . $spaces . "      for (::xml_schema::integer idx = 0; idx < x.get" .
                        ucfirst($fieldName) . "Count(); ++idx) {" .
                        "\n" . $spaces . "          this->add" . ucfirst($fieldName) . "(x.get" .
                        ucfirst($fieldName) . "(idx));" .
                        "\n" . $spaces . "      }" .
                        "\n" . $spaces . "    }" .
                        "\n" . $spaces . "    //else {" .
                        "\n" . $spaces . "    //  copyTraits<" . $fieldName .
                        "_type>::copyNeededFields(x.get" .
                        ucfirst($fieldName) . "(), this->" . $fieldName . "());" .
                        "\n" . $spaces . "    //}" .
                        "\n" . $spaces . "  }";
                }
                $line .= "\n" . $spaces . "}";

            }
        }

        # write the (possibly modified) line to the output file
        print FILEOUT "$line\n";
    }

    close FILEIN;
    close FILEOUT;

    rename $inputFile, "${inputFile}.orig";
    rename $outputFile, $inputFile;
}


# **********************************************************************
# Extra operation - add all classes to a standard registration method
# **********************************************************************

if (defined($opt_r) && keys(%classNames) > 0) {

    my $inputFile = $opt_r;
    my $outputFile = "${inputFile}.modified";

    open FILEIN, $inputFile or die "Unable to open input file $inputFile\n.";
    open FILEOUT, ">$outputFile" or die "Unable to open output file $outputFile\n";

    while (my $line = <FILEIN>) {
        chomp $line;

        # if this is the correct place to add include statements, do so now
        if ($line =~
            m/(\s*)(\S+)(\s+)dynamically generated includes end here(.*)/) {
            foreach my $fileName (@fileList) {
                if ($fileName =~ /.*inc[lude]*\/(\S+)h$/) {
                    if (defined($opt_p)) {
                        printf FILEOUT "#include <$opt_p/" . $1 . "h>\n";
                    }
                    else {
                        printf FILEOUT "#include <" . $1 . "h>\n";
                    }
                }
                elsif ($fileName =~ /h$/) {
                    if (defined($opt_p)) {
                        printf FILEOUT "#include <$opt_p/" .
                            basename($fileName) . ">\n";
                    }
                    else {
                        printf FILEOUT "#include <" .
                            basename($fileName) . ">\n";
                    }
                }
            }
        }

        # if this is the correct place to add registration messages, do so now
        if ($line =~ m/(\s*)(\S+)(\s+)dynamically generated registration lines end here(.*)/) {
            foreach my $className (sort(keys(%classNames))) {
                printf FILEOUT "\n";
                printf FILEOUT $1 . $className .
                    "::registerForDeserialization();\n";
            }
        }

        # write the input line to the output file
        print FILEOUT "$line\n";
    }

    close FILEIN;
    close FILEOUT;

    rename $outputFile, $inputFile;
}
