/* 
 * File:   DAQChannelMapBaseOLD.h
 * Author: denis
 *
 * This is the base class for DAQChannelMap.
 * It is an abstract class.
 * Here we precalculate some stuff, which are going to be use in the derived class
 *
 * Created on February 10, 2012, 10:17 AM
 */

#ifndef _DAQCHANNELMAPBASEOLD_H
#define	_DAQCHANNELMAPBASEOLD_H

#include "DAQChannelMap/DAQChannelMapErrorsOLD.h"
#include "DAQChannelMap/DAQChannelMapConstantsOLD.h"
#include "DAQChannelMap/DAQChannelMapConstants.h"

#include <string>
#include <vector>

typedef unsigned int uint;

namespace daqchannelmap {      //!< DAQ Channel Map Package

#ifndef EVEN
#define EVEN(x) (0 == ((x) % 2)) ? 1 : 0
#endif

#ifndef ODD
#define ODD(x)  (1 == ((x) % 2)) ? 1 : 0
#endif


//#define BOUNDS_CHECK_ALL /* Define this macro to enable bounds checking */

/// This thing has to have a comma at the end or it could be entirely empty
#define CHAN_IDX_0_1    0x0,

    class DAQChannelMapBaseOLD {

            /// Some type defs
            typedef std::vector< uint32_t >                              numberOfPlanesInBlock_t;
            typedef std::vector< uint32_t >                              numberOfPlanesInDiblock_t;
            typedef std::vector< numberOfPlanesInBlock_t   >             numberOfTypePlanesInBlock_t;
            typedef std::vector< numberOfPlanesInDiblock_t >             numberOfTypePlanesInDiblock_t;
            typedef std::vector< std::vector <uint32_t> >                numberOfPlanesInBlockDetector_t;
            typedef std::vector< std::vector <uint32_t> >                numberOfDCMsInDiblockDetector_t;



        public:
            DAQChannelMapBaseOLD();
            virtual ~DAQChannelMapBaseOLD();

            /// Print all detectors info
            bool    print() const;
            bool    print(const int detId, const offline_plane_t) const;


            /// Pixel number from APDPixel
            pixel_t         APDPixel2Pixel(apd_pixel_t) const;
            /// APDPixel number from Pixel
            apd_pixel_t     pixel2APDPixel(pixel_t) const;

            /// Diblock type given detectorID and diblock number
            DiBlock_TYPE    getDiBlockType(int detId      , diblock_t) const;
            
            /// Number of planes in the block given detectorID and block number
            uint32_t        getNumberOfPlanesInBlock  (int detId, block_t) const;
            /// Number of planes in the diblock given detectorID and diblock number
            uint32_t        getNumberOfPlanesInDiblock(int detId, diblock_t) const;

            /// Number of planes before block or diblock
            uint32_t        getNumberOfPlanesBeforeBlock(int detId, block_t) const;
            uint32_t        getNumberOfPlanesBeforeDiblock(int detId, diblock_t) const;
            
            /// Number of planes in the block given block type
            uint32_t        getNumberOfPlanesInBlock  (DetBlock_TYPE) const;
            /// Number of planes in the diblock given diblock type
            uint32_t        getNumberOfPlanesInDiblock(DiBlock_TYPE) const;

            /// Number of planes (X or Y) in the block given block type and type of plane(X or Y)
            uint32_t        getNumberOfPlaneTypeInBlock  (DetBlock_TYPE, DetPlane_TYPE) const;
            /// Number of planes (X or Y) in the diblock given diblock type and type of plane(X or Y)
            uint32_t        getNumberOfPlaneTypeInDiblock(DiBlock_TYPE , DetPlane_TYPE) const;

            /// Block type given diblock type an local block in diblock
            DetBlock_TYPE   getBlockType(DiBlock_TYPE, uint32_t local_block_number_in_diblock) const;
            /// Starting plane in the block given its type
            DetPlane_TYPE   getStartPlaneTypeInBlock(DetBlock_TYPE) const;

            /// get Plane type 
            DetPlane_TYPE   getPlaneType(const int detId, const offline_plane_t) const;
            //DetPlane_TYPE   getPlaneType(const int detId, const offline_plane_t, const DetBlock_TYPE block_type) const;

            /// Get minimum detector ID
            int             getDetectorIDMinimum() const;
            /// Get maximum detector ID
            int             getDetectorIDMaximum() const;

            /// The block number given the detectorID and offline plane number
            uint32_t        getBlockNumber            (int detId, offline_plane_t) const;
            /// The diblock number given the detectorID and offline plane number
            uint32_t        getDiblockNumber          (int detId, offline_plane_t) const;
            uint32_t        getDiblockNumber          (const block_t) const;
            uint32_t        getLocalBlockInDiblock    (const block_t) const;

            /// Local plane in block number given the detectorID and offline plane number
            offline_plane_t getLocalPlaneInBlock  (int detId, offline_plane_t) const;
            /// Local plane in diblock number given the detectorID and offline plane number
            offline_plane_t getLocalPlaneInDiblock(int detId, offline_plane_t) const;


            /// opposite plane view from the input. For instance Y = getOtherView(X)
            DetView_TYPE    getOtherView(DetView_TYPE) const;
            /// opposite plane view from the input. For instance Y = getOtherView(X)
            DetPlane_TYPE   getOtherPlaneType(DetPlane_TYPE) const;
            /// Conversion from DetPlane_TYPE to DetView_TYPE
            DetView_TYPE    detPlaneToDetView(DetPlane_TYPE) const;
            /// Conversion from DetView_TYPE to DetPlane_TYPE
            DetPlane_TYPE   detViewToDetPlane(DetView_TYPE) const;

            /// Number of DCMs for horizontal planes in the diblock given detectorID and Diblock type
            uint32_t        getNumberOfXDCM(int detId, DiBlock_TYPE) const;
            /// Number of DCMs for vertical planes in the diblock  given detectorID and Diblock type
            uint32_t        getNumberOfYDCM(int detId, DiBlock_TYPE) const;
            /// Total number of DCMs in the diblock  given detectorID and Diblock type
            uint32_t        getNumberOfDCM(int detId, DiBlock_TYPE) const;

            /// Number of channels in X plane for a given detectorID
            uint32_t        getNumberOfChannelsInXPlane(int detId) const;
            /// Number of channels in Y plane for a given detectorID
            uint32_t        getNumberOfChannelsInYPlane(int detId) const;

            /// Get Number of channels in plane
            uint32_t        getNumberOfChannelsInPlane(const int detId, const DetPlane_TYPE) const;
            uint32_t        getNumberOfChannelsInPlane(const int detId, const offline_plane_t) const;

            /// Number of cells per DCM for a given diblock type
            uint32_t        getNumberOfCellsPerDCM(DiBlock_TYPE) const;

            /// Total number of blocks for a given detector
            uint32_t        getTotalNumberOfBlocks(int detId) const; ///< Total number of diblocks in the detector
            /// Total number of diblocks for a given detector
            uint32_t        getTotalNumberOfDiblocks(int detId) const; ///< Total number of diblocks in the detector

            /// Total number of Planes in the detector
            uint32_t        getTotalNumberOfPlanes(const int detId) const;

            // Error Message Related
            const char* errorString(DAQChanMapErrorCodes err_code) const;

        private:
            
            uint32_t        getBlockOrDiblock            (int detId
                                                         ,offline_plane_t
                                                         ,const numberOfPlanesInBlockDetector_t&
                                                         ,const BlockNumerationOffset
                                                          ) const;
            uint32_t        getBlockOrDiblock            (int detId
                                                         ,offline_plane_t
                                                         ,const uint32_t max_number_of_blocks_or_diblocks
                                                         ,const uint**
                                                         ,const BlockNumerationOffset
                                                         ) const;

            offline_plane_t getLocalPlaneInBlockOrDiblock(int detId
                                                         ,offline_plane_t
                                                         ,const numberOfPlanesInBlockDetector_t&
                                                         ,const BlockNumerationOffset
                                                         ) const;

            offline_plane_t getLocalPlaneInBlockOrDiblock(const int detId
                                                         ,const offline_plane_t
                                                         ,const uint32_t max_number_of_blocks_or_diblocks
                                                         ,const uint**
                                                         ,const BlockNumerationOffset
                                                         ) const;

            /// Initialization methods
            bool            initNumberOfCellsPerDCMInDiblock();
            bool            initBlockOrderInDiblock();
            bool            initPlanesInBlockID();
            bool            initNumberOfTypePlanesInBlock();
            bool            initPlanesInDiblockID();
            bool            initStartPlaneTypeInBlock();
            bool            initPlanesInBlockAndDiblockDetector();
            bool            initNumberOfDCMsInDiblockDetector();
            bool            initPlaneTypePlaneNumber();
            bool            initDetectorDiBlockConfig();


            /// Checks for Initialization methods
            bool            isinitNumberOfCellsPerDCMInDiblock();
            bool            isinitBlockOrderInDiblock();
            bool            isinitPlanesInBlockID();
            bool            isinitNumberOfTypePlanesInBlock();
            bool            isinitPlanesInDiblockID();
            bool            isinitStartPlaneTypeInBlock();
            bool            isinitPlanesInBlockAndDiblockDetector();
            bool            isinitNumberOfDCMsInDiblockDetector();
            bool            isinitDetectorDiBlockConfig();

            bool            initApdPixel2PixelMap();

            bool            allocateArray(uint32_t**&, const std::vector< std::vector<uint32_t> >&);
            bool            allocateSize (uint32_t*& , const std::vector< std::vector<uint32_t> >&);
        private:
                static const char*         DetectorLabels[];
                const static unsigned int  NUM_XCHAN_PLANE[];
                const static unsigned int  NUM_YCHAN_PLANE[];
                const static unsigned int  NUM_DIBLOCK_DET[];
                const static unsigned int  NUM_DCMS_DIBLOCK[];
                const static unsigned int  NUM_X_DCMS[];
                const static unsigned int  NUM_Y_DCMS[];
                const static unsigned int  apd_pixel2cell_v0[],apd_cell2pixel_v0[];

		const static unsigned int  DetectorDiBlockConfig_IPND[];
                const static unsigned int  DetectorDiBlockConfig_NDOS[];
                const static unsigned int  DetectorDiBlockConfig_NEARDET[];
                const static unsigned int  DetectorDiBlockConfig_FARDET[];
                const static unsigned int  DetectorDiBlockConfig_TESTBEAM_2Block_1x1[];
                const static unsigned int  DetectorDiBlockConfig_TESTBEAM_3Block_1x1[];
                const static unsigned int  DetectorDiBlockConfig_TESTBEAM_2Block_2x2[];
                const static unsigned int  DetectorDiBlockConfig_TESTBEAM_3Block_2x2[];
                std::vector< std::vector <uint32_t> > DetectorDiBlockConfigVector;
                uint32_t**                            DetectorDiBlockConfigArray;

                std::vector<uint32_t> getVector(const unsigned int*, const unsigned int maxsize=1000000);
                void printDiBlockConfigVector();


                std::vector<uint32_t>           _planesInDiblockType; //Number of planes in diblock type
                std::vector<uint32_t>           _planesInBlockType;   //Number of planes in diblock type
                uint32_t*                       _planesInDiblockTypeArray;
                uint32_t*                       _planesInBlockTypeArray;

                numberOfPlanesInBlockDetector_t _planesInDiblockDetector; // array depends on detector and diblock number
                numberOfPlanesInBlockDetector_t _planesInBlockDetector; // array depends on detector and diblock number
                unsigned int**                  _planesInDiblockDetectorArray;
                unsigned int**                  _planesInBlockDetectorArray;
                unsigned int*                   _planesInBlockDetectorSize;
                unsigned int*                   _planesInDiblockDetectorSize;

                
                numberOfPlanesInBlockDetector_t _planesBeforeDiblockDetector; // array depends on detector and diblock number
                numberOfPlanesInBlockDetector_t _planesBeforeBlockDetector;   // array depends on detector and diblock number
                unsigned int**                  _planesBeforeDiblockDetectorArray;
                unsigned int**                  _planesBeforeBlockDetectorArray;

                numberOfPlanesInBlockDetector_t _BlockOrderInDiblock; // The order of blocks in the diblock
                unsigned int**                  _BlockOrderInDiblockArray;
                

                std::vector<DetPlane_TYPE>      _startPlaneTypeInBlock;  ///< Does the block start with X or Y plane. Keyed by block
                DetPlane_TYPE*                  _startPlaneTypeInBlockArray;

                numberOfTypePlanesInBlock_t     _numberOfTypePlanesInBlock;   /// Keyed bu the DetPlane_TYPE and
                numberOfTypePlanesInDiblock_t   _numberOfTypePlanesInDiBlock; /// Keyed bu the DetPlane_TYPE and
                uint32_t**                      _numberOfTypePlanesInBlockArray;   /// Keyed bu the DetPlane_TYPE and
                uint32_t**                      _numberOfTypePlanesInDiBlockArray; /// Keyed bu the DetPlane_TYPE and

                std::vector<uint32_t>           _numberOfCellsPerDCMInDiblock; /// Number of cells that each DCM caries for each diblock type
                uint32_t*                       _numberOfCellsPerDCMInDiblockArray;

                numberOfDCMsInDiblockDetector_t _numberOfXDCMsInDiblockDetector;// holds how many DCMs we make in a diblock of particular type and of particular detector
                numberOfDCMsInDiblockDetector_t _numberOfYDCMsInDiblockDetector;// holds how many DCMs we make in a diblock of particular type and of particular detector
                uint32_t**                      _numberOfXDCMsInDiblockDetectorArray;
                uint32_t**                      _numberOfYDCMsInDiblockDetectorArray;

                std::vector <pixel_t>           _apdPixel2CellMap;
                std::vector <apd_pixel_t>       _cell2ApdPixelMap;
                pixel_t*                        _apdPixel2CellMapArray;
                apd_pixel_t*                    _cell2ApdPixelMapArray;
                

                std::vector < std::vector <uint32_t> > _planeType_planeNumber; /// Plane type as a function of offline plane number and DetectorID
                uint32_t**                             _planeType_planeNumberArray;
                
                std::vector <uint32_t>          _totalNumberOfPlanes;


    };
}// end of namespace daqchannelmap

#endif	/* _DAQCHANNELMAPBASE_H */

