#ifndef DAQDATAFORMATS_H
#define DAQDATAFORMATS_H

/*!

  The NOVA DAQDataFormats Library contains classes for all of the
  major data types that are produced by the data acquisition hardware
  and software chains.

  DAQDataFormats.h should be included when ever be included whenever
  the programmer needs access to forward declarations of these objects
  (i.e. in other class headers)

  For access to the individual class objects, it is safe to include 
  both this library header, and the header for the object you need to
  access (i.e. include "DAQDataFormats.h"  and include "RawNanoSlice.h"

  anorman@fnal.gov 
*/

namespace daqdataformats {

// Forward declarations of library classes
class RawDAQData;		//! Base Class for all DAQData Formats

// NanoSlice Formats
class RawNanoSlice;		//! Class for single hit samples

// MicroSlice Formats
class RawMicroSlice;		//! Class for the general MicroSlice
class RawMicroSliceHeader;	//! Class for the header on the MicroSlice block
class RawTimingMarker;          //! Class for the timing marker on the MicroSlice block  

// Milliscile Formats
class RawMilliSlice;		//! Class for the general MicroSlice
class RawMilliSliceHeader;	//! Class for the header on the MilliSlice block
class RawMilliDCMChan;   	//! Class for the DCM Channel identification fields
class RawMilliSubframeHeader;	//! Class for the Subframe Headers
class RawMilliSubframe;		//! Class for the Subframe
class RawMilliSliceIndex;       //! Class for the MilliSlice Index


// Trigger Formats
class RawTrigger;		//! Class for the general trigger block
class RawTriggerHeader;		//! Class for the header in the trigger block
class RawTriggerTimingMarker;	//! Class for the timing marker in the trigger block
class RawTriggerRange;		//! Class for the range fields in the trigger block
class RawTriggerTime;		//! Class for the trigger's generation time
class RawTriggerMask;		//! Class for the trigger bit masks

// Event Formats
class RawEvent;
class RawEventHeader;
class RawEventTail;


// Simulatolastses
class FEBSimulator;
class DCMSimulator;

}
#endif /* DAQDATAFORMATS_H */
