/* 
 * File:   Macros.h
 * Author: denis
 *
 * This file contains macros that are useful for DAQDataFormats initializations and such
 * Most of these macros should be used in the Header files only
 *
 * Created on February 19, 2012, 10:38 AM
 */

#ifndef MACROS_H
#define	MACROS_H

#include "DAQDataFormats/FunctionBind.h"

//---------------------------------------------------------------------------------
/// These macros are used to put definitions of function pointers in the Header of the first version of the class
/// Put them in the Header!!!!!!!!!!!!!!!!!!!
//---------------------------------------------------------------------------------

/// Define Get function pointer. Put it in the header of the first class. Custom output units
#define INIT_CUSTOM_DELEGATE(FULLFUNCTIONNAME, OUTPUT, ...) protected: DELEGATE(CURRENT_CLASS, _function_##FULLFUNCTIONNAME, OUTPUT,__VA_ARGS__ )

/// Define Get function pointer. Put it in the header of the first class. Custom output units
#define INIT_CUSTOM_GET_FUNCTION(FULLFUNCTIONNAME, OUTPUT_UNITS) INIT_CUSTOM_DELEGATE(FULLFUNCTIONNAME, OUTPUT_UNITS) const;

/// Define Set function pointer. Put it in the header of the first class. Custom input units
#define INIT_CUSTOM_SET_FUNCTION(FULLFUNCTIONNAME, INPUT_UNITS)  INIT_CUSTOM_DELEGATE(FULLFUNCTIONNAME, bool, INPUT_UNITS);

/// Define both Get and Set function pointers. Put it in the header of the first class. Custom units
#define INIT_GET_AND_SET_FUNCTION(FUNCT, UNITS)        \
            INIT_CUSTOM_GET_FUNCTION(get##FUNCT, UNITS) \
            INIT_CUSTOM_SET_FUNCTION(set##FUNCT, UNITS)

/// Define both Is and Set function pointers. Put it in the header of the first class
#define INIT_IS_AND_SET_FUNCTION(FUNCT)                \
            INIT_CUSTOM_GET_FUNCTION(is##FUNCT , bool) \
            INIT_CUSTOM_SET_FUNCTION(set##FUNCT, bool)

//---------------------------------------------------------------------------------
/// These macros are used to define and implement get and set methods in Main Class
/// Put them in the Header!!!!!!!!!!!!!!!!!!!
//---------------------------------------------------------------------------------

/// Checking of the binding function if it's defined
#define CHECK_BINDING_FUNCTION(function) if (! function){                                                  \
                                            std::cout<<"Error!!! Function "<<__PRETTY_FUNCTION__<<" "      \
                                                     <<"is not binded. For version "<<_version<<". "       \
                                                     <<"Check the appropriate setFunctionPointers method." \
                                                     <<std::endl;                                          \
                                            exit(1);                                                       \
                                         }
/// Execute and return the delegate
#define EXECUTE_FUNCTION(FULLFUNCTIONNAME)                                             \
                EXECUTE_GENERAL_FUNCTION(_function_##FULLFUNCTIONNAME)

#define RETURN_FUNCTION(FULLFUNCTIONNAME)                                              \
                EXECUTE_ON_DEBUG(CHECK_BINDING_FUNCTION(_function_##FULLFUNCTIONNAME)) \
                return EXECUTE_GENERAL_FUNCTION(_function_##FULLFUNCTIONNAME)

//------------------------------------------------------------------------------
/// These macros are used to implement gets and sets of the fields
//------------------------------------------------------------------------------
#define IMPLEMENT_GENERAL_GET_IN_FIRST_VERSION_CLASS(PROTECTIVENESS, NAMESPACE, FUNC, UNITS)  \
PROTECTIVENESS: UNITS get##FUNC() const {                                                     \
    const uint32_t* pointer = &(RAWBUFF32[NAMESPACE::WORD_##FUNC]);                           \
    return (UNITS) (BITFIELDGET(*pointer, NAMESPACE::MASK_##FUNC, NAMESPACE::SHIFT_##FUNC));  \
}
#define IMPLEMENT_GENERAL_GET_OR_IS_IN_FIRST_VERSION_CLASS_BOOL(PROTECTIVENESS, PREFUNCT, NAMESPACE, FUNC) \
PROTECTIVENESS: bool PREFUNCT##FUNC() const {                                                              \
    const uint32_t* pointer = &(RAWBUFF32[NAMESPACE::WORD_##FUNC]);                                        \
    return uint32_tToBool(BITFIELDGET(*pointer, NAMESPACE::MASK_##FUNC, NAMESPACE::SHIFT_##FUNC));         \
}
#define IMPLEMENT_GENERAL_SET_IN_FIRST_VERSION_CLASS(PROTECTIVENESS, NAMESPACE, FUNC, UNITS)  \
PROTECTIVENESS: bool  set##FUNC(const UNITS value) {                                          \
    uint32_t* pointer = &(RAWBUFF32[NAMESPACE::WORD_##FUNC]);                                 \
    BITFIELDSET(*pointer, value, NAMESPACE::MASK_##FUNC, NAMESPACE::SHIFT_##FUNC);            \
    return true;                                                                              \
}
#define IMPLEMENT_GENERAL_SET_IN_FIRST_VERSION_CLASS_BOOL(PROTECTIVENESS, NAMESPACE, FUNC)    \
PROTECTIVENESS: bool  set##FUNC(const bool value) {                                           \
    uint32_t*       pointer = &(RAWBUFF32[NAMESPACE::WORD_##FUNC]);                           \
    const uint32_t  val     = boolToUint32_t(value);                                          \
    BITFIELDSET(*pointer, val, NAMESPACE::MASK_##FUNC, NAMESPACE::SHIFT_##FUNC);              \
    return true;                                                                              \
}

/// Implement Get, Set method with namespace and units. And then with units=bool
#define IMPLEMENT_GET_IN_FIRST_VERSION_CLASS(NAMESPACE, FUNC, UNITS)  IMPLEMENT_GENERAL_GET_IN_FIRST_VERSION_CLASS(public, NAMESPACE, FUNC, UNITS)
#define IMPLEMENT_SET_IN_FIRST_VERSION_CLASS(NAMESPACE, FUNC, UNITS)  IMPLEMENT_GENERAL_SET_IN_FIRST_VERSION_CLASS(public, NAMESPACE, FUNC, UNITS)
#define IMPLEMENT_GET_IN_FIRST_VERSION_CLASS_BOOL(NAMESPACE, FUNC)    IMPLEMENT_GENERAL_GET_OR_IS_IN_FIRST_VERSION_CLASS_BOOL(public, get, NAMESPACE, FUNC)
#define IMPLEMENT_SET_IN_FIRST_VERSION_CLASS_BOOL(NAMESPACE, FUNC)    IMPLEMENT_GENERAL_SET_IN_FIRST_VERSION_CLASS_BOOL(public, NAMESPACE, FUNC)
#define IMPLEMENT_IS_IN_FIRST_VERSION_CLASS(NAMESPACE, FUNC)          IMPLEMENT_GENERAL_GET_OR_IS_IN_FIRST_VERSION_CLASS_BOOL(public, is, NAMESPACE, FUNC)

#define IMPLEMENT_GET_IN_FIRST_VERSION_CLASS_PROTECTED(NAMESPACE, FUNC, UNITS)  IMPLEMENT_GENERAL_GET_IN_FIRST_VERSION_CLASS(protected, NAMESPACE, FUNC, UNITS)
#define IMPLEMENT_SET_IN_FIRST_VERSION_CLASS_PROTECTED(NAMESPACE, FUNC, UNITS)  IMPLEMENT_GENERAL_SET_IN_FIRST_VERSION_CLASS(protected, NAMESPACE, FUNC, UNITS)
#define IMPLEMENT_GET_IN_FIRST_VERSION_CLASS_BOOL_PROTECTED(NAMESPACE, FUNC)    IMPLEMENT_GENERAL_GET_OR_IS_IN_FIRST_VERSION_CLASS_BOOL(protected, get, NAMESPACE, FUNC)
#define IMPLEMENT_SET_IN_FIRST_VERSION_CLASS_BOOL_PROTECTED(NAMESPACE, FUNC)    IMPLEMENT_GENERAL_SET_IN_FIRST_VERSION_CLASS(protected, NAMESPACE, FUNC)
#define IMPLEMENT_IS_IN_FIRST_VERSION_CLASS_PROTECTED(NAMESPACE, FUNC)          IMPLEMENT_GENERAL_GET_OR_IS_IN_FIRST_VERSION_CLASS_BOOL(protected, is, NAMESPACE, FUNC)

#define IMPLEMENT_GET(FUNC, UNITS) IMPLEMENT_GET_IN_FIRST_VERSION_CLASS(VERSION_NAMESPACE::CURRENT_FIELDS_SUB_NAMESPACE, FUNC, UNITS)
#define IMPLEMENT_SET(FUNC, UNITS) IMPLEMENT_SET_IN_FIRST_VERSION_CLASS(VERSION_NAMESPACE::CURRENT_FIELDS_SUB_NAMESPACE, FUNC, UNITS)
#define IMPLEMENT_GET_BOOL(FUNC)   IMPLEMENT_GET_IN_FIRST_VERSION_CLASS_BOOL(VERSION_NAMESPACE::CURRENT_FIELDS_SUB_NAMESPACE, FUNC)
#define IMPLEMENT_SET_BOOL(FUNC)   IMPLEMENT_SET_IN_FIRST_VERSION_CLASS_BOOL(VERSION_NAMESPACE::CURRENT_FIELDS_SUB_NAMESPACE, FUNC)
#define IMPLEMENT_IS(FUNC)         IMPLEMENT_IS_IN_FIRST_VERSION_CLASS(VERSION_NAMESPACE::CURRENT_FIELDS_SUB_NAMESPACE, FUNC)

#define IMPLEMENT_GET_NOVERSION(FUNC, UNITS) IMPLEMENT_GET_IN_FIRST_VERSION_CLASS(CURRENT_FIELDS_SUB_NAMESPACE, FUNC, UNITS)
#define IMPLEMENT_SET_NOVERSION(FUNC, UNITS) IMPLEMENT_SET_IN_FIRST_VERSION_CLASS(CURRENT_FIELDS_SUB_NAMESPACE, FUNC, UNITS)
#define IMPLEMENT_GET_NOVERSION_BOOL(FUNC)   IMPLEMENT_GET_IN_FIRST_VERSION_CLASS_BOOL(CURRENT_FIELDS_SUB_NAMESPACE, FUNC)
#define IMPLEMENT_SET_NOVERSION_BOOL(FUNC)   IMPLEMENT_SET_IN_FIRST_VERSION_CLASS_BOOL(CURRENT_FIELDS_SUB_NAMESPACE, FUNC)
#define IMPLEMENT_IS_NOVERSION(FUNC)         IMPLEMENT_IS_IN_FIRST_VERSION_CLASS(CURRENT_FIELDS_SUB_NAMESPACE, FUNC)

#define IMPLEMENT_GET_NOVERSION_PROTECTED(FUNC, UNITS) IMPLEMENT_GET_IN_FIRST_VERSION_CLASS_PROTECTED(CURRENT_FIELDS_SUB_NAMESPACE, FUNC, UNITS)
#define IMPLEMENT_SET_NOVERSION_PROTECTED(FUNC, UNITS) IMPLEMENT_SET_IN_FIRST_VERSION_CLASS_PROTECTED(CURRENT_FIELDS_SUB_NAMESPACE, FUNC, UNITS)
#define IMPLEMENT_GET_NOVERSION_BOOL_PROTECTED(FUNC)   IMPLEMENT_GET_IN_FIRST_VERSION_CLASS_BOOL_PROTECTED(CURRENT_FIELDS_SUB_NAMESPACE, FUNC)
#define IMPLEMENT_SET_NOVERSION_BOOL_PROTECTED(FUNC)   IMPLEMENT_SET_IN_FIRST_VERSION_CLASS_BOOL_PROTECTED(CURRENT_FIELDS_SUB_NAMESPACE, FUNC, UNITS)
#define IMPLEMENT_IS_NOVERSION_PROTECTED(FUNC)         IMPLEMENT_IS_IN_FIRST_VERSION_CLASS_PROTECTED(CURRENT_FIELDS_SUB_NAMESPACE, FUNC)


#define IMPLEMENT_GET_AND_SET(FUNC, UNITS) \
            IMPLEMENT_GET(FUNC, UNITS)     \
            IMPLEMENT_SET(FUNC, UNITS)
#define IMPLEMENT_GET_AND_SET_BOOL(FUNC) \
            IMPLEMENT_GET_BOOL(FUNC)     \
            IMPLEMENT_SET_BOOL(FUNC)
#define IMPLEMENT_IS_AND_SET(FUNC)       \
            IMPLEMENT_IS(FUNC)           \
            IMPLEMENT_SET_BOOL(FUNC)

#define IMPLEMENT_GET_AND_SET_NOVERSION(FUNC, UNITS) \
            IMPLEMENT_GET_NOVERSION(FUNC, UNITS)     \
            IMPLEMENT_SET_NOVERSION(FUNC, UNITS)
#define IMPLEMENT_GET_AND_SET_NOVERSION_BOOL(FUNC) \
            IMPLEMENT_GET_NOVERSION_BOOL(FUNC)     \
            IMPLEMENT_SET_NOVERSION_BOOL(FUNC)
#define IMPLEMENT_IS_AND_SET_NOVERSION(FUNC) \
            IMPLEMENT_IS_NOVERSION(FUNC)     \
            IMPLEMENT_SET_NOVERSION_BOOL(FUNC)

#define IMPLEMENT_GET_AND_SET_NOVERSION_PROTECTED(FUNC, UNITS) \
            IMPLEMENT_GET_NOVERSION_PROTECTED(FUNC, UNITS)     \
            IMPLEMENT_SET_NOVERSION_PROTECTED(FUNC, UNITS)
#define IMPLEMENT_GET_AND_SET_NOVERSION_BOOL_PROTECTED(FUNC) \
            IMPLEMENT_GET_NOVERSION_BOOL_PROTECTED(FUNC)     \
            IMPLEMENT_SET_NOVERSION_BOOL_PROTECTED(FUNC)
#define IMPLEMENT_IS_AND_SET_NOVERSION_PROTECTED(FUNC) \
            IMPLEMENT_IS_NOVERSION_PROTECTED(FUNC)     \
            IMPLEMENT_SET_NOVERSION_BOOL_PROTECTED(FUNC)

//------------------------------------------------------------------------------
#define DEFINE_SHIFT_FROM_MASK(FUNC) static const uint32_t SHIFT_##FUNC = daqdataformats::calculateShiftFromAGivenMask(MASK_##FUNC);

//------------------------------------------------------------------------------
#define IMPLEMENT_FIRST_GET(FUNC, UNITS)                         \
        IMPLEMENT_GET(FUNC, UNITS);                              \
        INIT_CUSTOM_DELEGATE(get##FUNC, UNITS, ) const;
#define IMPLEMENT_FIRST_SET(FUNC, UNITS)                         \
            IMPLEMENT_SET(FUNC, UNITS);                          \
            INIT_CUSTOM_DELEGATE(set##FUNC, bool , const UNITS);
#define IMPLEMENT_FIRST_GET_BOOL(FUNC)                           \
        IMPLEMENT_GET_BOOL(FUNC);                                \
        INIT_CUSTOM_DELEGATE(get##FUNC, bool, ) const;
#define IMPLEMENT_FIRST_SET_BOOL(FUNC)                           \
            IMPLEMENT_SET_BOOL(FUNC);                            \
            INIT_CUSTOM_DELEGATE(set##FUNC, bool , const bool);
#define IMPLEMENT_FIRST_IS(FUNC)                                 \
        IMPLEMENT_IS(FUNC);                                      \
        INIT_CUSTOM_DELEGATE(is##FUNC, bool, ) const;


#define IMPLEMENT_FIRST_GET_AND_SET(FUNC, UNITS)\
            IMPLEMENT_FIRST_GET(FUNC, UNITS)    \
            IMPLEMENT_FIRST_SET(FUNC, UNITS)
#define IMPLEMENT_FIRST_GET_AND_SET_BOOL(FUNC)  \
            IMPLEMENT_FIRST_GET_BOOL(FUNC);     \
            IMPLEMENT_FIRST_SET_BOOL(FUNC);
#define IMPLEMENT_FIRST_IS_AND_SET(FUNC)     \
            IMPLEMENT_FIRST_IS(FUNC);        \
            IMPLEMENT_FIRST_SET_BOOL(FUNC);

#define IMPLEMENT_FUNCTION_RETURNING_FUNCTION_POINTER_0PAR(FUNC, CONSTNESS, OUTPUT) \
            OUTPUT FUNC() CONSTNESS {RETURN_FUNCTION(FUNC)();}
#define IMPLEMENT_FUNCTION_RETURNING_FUNCTION_POINTER_1PAR(FUNC, CONSTNESS, OUTPUT, INP1) \
            OUTPUT FUNC(INP1 a) CONSTNESS {RETURN_FUNCTION(FUNC)(a);}
#define IMPLEMENT_FUNCTION_RETURNING_FUNCTION_POINTER_2PAR(FUNC, CONSTNESS, OUTPUT, INP1,INP2) \
            OUTPUT FUNC(INP1 a,INP2 b) CONSTNESS {RETURN_FUNCTION(FUNC)(a,b);}
#define IMPLEMENT_FUNCTION_RETURNING_FUNCTION_POINTER_3PAR(FUNC, CONSTNESS, OUTPUT, INP1,INP2,INP3) \
            OUTPUT FUNC(INP1 a,INP2 b,INP3 c) CONSTNESS {RETURN_FUNCTION(FUNC)(a, b, c);}
#define IMPLEMENT_FUNCTION_RETURNING_FUNCTION_POINTER_4PAR(FUNC, CONSTNESS, OUTPUT, INP1,INP2,INP3,INP4) \
            OUTPUT FUNC(INP1 a,INP2 b,INP3 c,INP4 d) CONSTNESS {RETURN_FUNCTION(FUNC)(a,b,c,d);}
#define IMPLEMENT_FUNCTION_RETURNING_FUNCTION_POINTER_5PAR(FUNC, CONSTNESS, OUTPUT, INP1,INP2,INP3,INP4,INP5) \
            OUTPUT FUNC(INP1 a,INP2 b,INP3 c,INP4 d,INP5 e) CONSTNESS {RETURN_FUNCTION(FUNC)(a,b,c,d,e);}

#define IMPLEMENT_VOID_RETURNING_FUNCTION_POINTER_0PAR(FUNC, CONSTNESS) \
            void FUNC() CONSTNESS {EXECUTE_FUNCTION(FUNC)();}
#define IMPLEMENT_VOID_RETURNING_FUNCTION_POINTER_1PAR(FUNC, CONSTNESS, INP1) \
            void FUNC(INP1 a) CONSTNESS {EXECUTE_FUNCTION(FUNC)(a);}
#define IMPLEMENT_VOID_RETURNING_FUNCTION_POINTER_2PAR(FUNC, CONSTNESS, INP1,INP2) \
            void FUNC(INP1 a,INP2 b) CONSTNESS {EXECUTE_FUNCTION(FUNC)(a,b);}
#define IMPLEMENT_VOID_RETURNING_FUNCTION_POINTER_3PAR(FUNC, CONSTNESS, INP1,INP2,INP3) \
            void FUNC(INP1 a,INP2 b,INP3 c) CONSTNESS {EXECUTE_FUNCTION(FUNC)(a, b, c);}
#define IMPLEMENT_VOID_RETURNING_FUNCTION_POINTER_4PAR(FUNC, CONSTNESS, INP1,INP2,INP3,INP4) \
            void FUNC(INP1 a,INP2 b,INP3 c,INP4 d) CONSTNESS {EXECUTE_FUNCTION(FUNC)(a,b,c,d);}
#define IMPLEMENT_VOID_RETURNING_FUNCTION_POINTER_5PAR(FUNC, CONSTNESS, INP1,INP2,INP3,INP4,INP5) \
            void FUNC(INP1 a,INP2 b,INP3 c,INP4 d,INP5 e) CONSTNESS {EXECUTE_FUNCTION(FUNC)(a,b,c,d,e);}

//------------------------------------------------------------------------------
#define IMPLEMENT_GET_MAIN(FUNC, UNITS)\
            IMPLEMENT_FUNCTION_RETURNING_FUNCTION_POINTER_0PAR(get##FUNC, const, UNITS); 
#define IMPLEMENT_SET_MAIN(FUNC, UNITS)\
            IMPLEMENT_FUNCTION_RETURNING_FUNCTION_POINTER_1PAR(set##FUNC,      , bool, const UNITS);
#define IMPLEMENT_IS_MAIN(FUNC)\
            IMPLEMENT_FUNCTION_RETURNING_FUNCTION_POINTER_0PAR(is##FUNC , const, bool);



#define IMPLEMENT_GET_AND_SET_MAIN(FUNC, UNITS) \
            IMPLEMENT_GET_MAIN(FUNC, UNITS)     \
            IMPLEMENT_SET_MAIN(FUNC, UNITS)
#define IMPLEMENT_IS_AND_SET_MAIN(FUNC)   \
            IMPLEMENT_IS_MAIN(FUNC)       \
            IMPLEMENT_SET_MAIN(FUNC, bool)

#define IMPLEMENT_MAIN_init            IMPLEMENT_VOID_RETURNING_FUNCTION_POINTER_0PAR    (init           ,      )
#define IMPLEMENT_MAIN_print           void print(std::ostream& os=std::cout) const {EXECUTE_FUNCTION(print)(os);}
#define IMPLEMENT_MAIN_sizeofdata      IMPLEMENT_FUNCTION_RETURNING_FUNCTION_POINTER_0PAR(sizeofdata     , const, uint32_t)
#define IMPLEMENT_MAIN_readData        IMPLEMENT_FUNCTION_RETURNING_FUNCTION_POINTER_1PAR(readData       ,      , void*, const void*)
#define IMPLEMENT_MAIN_setBufferSource IMPLEMENT_FUNCTION_RETURNING_FUNCTION_POINTER_1PAR(setBufferSource,      , bool , const void*)

#define IMPLEMENT_MAIN_STANDARD                          \
            IMPLEMENT_MAIN_init                          \
            using PREVIOUS_VERSION_CLASS_NAME::print;    \
            IMPLEMENT_MAIN_print                         \
            IMPLEMENT_MAIN_sizeofdata                    \
            using PREVIOUS_VERSION_CLASS_NAME::readData; \
            IMPLEMENT_MAIN_readData

//------------------------------------------------------------------------------
#define INIT_FIRST(FUNC, CONSTNESS, OUTPUT_UNITS, ...)                       \
            public: OUTPUT_UNITS FUNC(__VA_ARGS__) CONSTNESS;                \
            INIT_CUSTOM_DELEGATE(FUNC, OUTPUT_UNITS,__VA_ARGS__) CONSTNESS;

#define INIT_FIRST_GET(FUNC, UNITS) INIT_FIRST(get##FUNC, const, UNITS, )
#define INIT_FIRST_SET(FUNC, UNITS) INIT_FIRST(set##FUNC,      , bool, const UNITS)
#define INIT_FIRST_GET_AND_SET(FUNC, UNITS)	\
  INIT_FIRST_GET(FUNC, UNITS)				\
  INIT_FIRST_SET(FUNC, UNITS) 

//------------------------------------------------------------------------------
#define BIND_FUNCTION_TO_CUSTOM_FUNCTION(FUNC,CLASS,FIRSTVERSION_NAMESPACE,BINDING,CONSTNESS,OUTPUT,...) \
  _function_##FUNC = (DELEGATE(FIRSTVERSION_NAMESPACE::CLASS, , OUTPUT, __VA_ARGS__) CONSTNESS)&BINDING

#define BIND_FUNCTION(FUNC,CLASS,FIRSTVERSION_NAMESPACE,BINDINGVERSION_NAMESPACE,CONSTNESS,OUTPUT,...) \
  BIND_FUNCTION_TO_CUSTOM_FUNCTION(FUNC,CLASS,FIRSTVERSION_NAMESPACE,BINDINGVERSION_NAMESPACE::CLASS::FUNC,CONSTNESS,OUTPUT,__VA_ARGS__)


#define BIND_CUSTOM_FUNCTION(FUNC, FIRSTVERSION_NAMESPACE, CONSTNESS, OUTPUT,...) BIND_FUNCTION(FUNC, CURRENT_CLASS, FIRSTVERSION_NAMESPACE, CURRENT_VERSION_NAMESPACE, CONSTNESS, OUTPUT, __VA_ARGS__)


#define BIND_GET(FUNC, FIRSTVERSION_NAMESPACE, OUTPUT) BIND_FUNCTION(get##FUNC,CURRENT_CLASS,FIRSTVERSION_NAMESPACE,CURRENT_VERSION_NAMESPACE, const, OUTPUT, )
#define BIND_SET(FUNC, FIRSTVERSION_NAMESPACE, INPUT)  BIND_FUNCTION(set##FUNC,CURRENT_CLASS,FIRSTVERSION_NAMESPACE,CURRENT_VERSION_NAMESPACE,      , bool  , const INPUT)
#define BIND_IS(FUNC , FIRSTVERSION_NAMESPACE)         BIND_FUNCTION(is##FUNC ,CURRENT_CLASS,FIRSTVERSION_NAMESPACE,CURRENT_VERSION_NAMESPACE, const, bool, )

#define BIND_GET_AND_SET(FUNC, FIRSTVERSION_NAMESPACE, OUTPUT)	\
  BIND_GET(FUNC, FIRSTVERSION_NAMESPACE, OUTPUT);			\
  BIND_SET(FUNC, FIRSTVERSION_NAMESPACE, OUTPUT);

#define BIND_IS_AND_SET(...)            \
            BIND_IS(__VA_ARGS__);       \
            BIND_SET(__VA_ARGS__, bool);


/// Binding standard functions
#define BIND_STADARD_FUNCTION(FUNC) _function_##FUNC = (FUNC##_t) &CURRENT_VERSION_NAMESPACE::CURRENT_CLASS::FUNC;

#define BIND_STADARD_FUNCTIONS                      \
            BIND_STADARD_FUNCTION(init);            \
            BIND_STADARD_FUNCTION(print);           \
            BIND_STADARD_FUNCTION(setBufferSource); \
            BIND_STADARD_FUNCTION(readData);                           
        


//------------------------------------------------------------------------------
//
#define IMPLEMENT_FIGUREOUT_VERSION_STATIC_WITH_CUSTOM_GET_VERSION(CLASS, GETVERSIONFUNCTION)                                                                     \
            version_t        figureOutTheVersion      (const void* buffer) const {return figureOutTheVersionStatic(buffer);} \
            static version_t figureOutTheVersionStatic(const void* buffer){                                           \
                CLASS header_zero((void*)buffer, 0);                                                                  \
                return (version_t)header_zero.GETVERSIONFUNCTION();                                                   \
            }


/// This should be in the header
#define IMPLEMENT_FIGUREOUT_VERSION_STATIC(CLASS) IMPLEMENT_FIGUREOUT_VERSION_STATIC_WITH_CUSTOM_GET_VERSION(CLASS, getVersion)

#endif	/* MACROS_H */

