#ifndef NOVA_CHECKSUM_H
#define NOVA_CHECKSUM_H

#include <boost/crc.hpp> ///< Boost library for CRC

/// CRC-32C (Castagnoli) iSCSI polynomial (See A.Norman, NOvA-doc-DB-4619)
typedef boost::crc_optimal<32, 0x1EDC6F41, 0x0, 0x0, true, true> nova_crc_32_type;

namespace daqdataformats {
    class NOvACheckSum {

        public:
            static NOvACheckSum* getInstance(); //!< Method to obtain the class instance
            ~NOvACheckSum();                    //!< Default Destructor.

            uint32_t processCRC(void* buffer, uint32_t size);

        private:
            NOvACheckSum();//!< Default Constructor. Has to be private, since we are using singleton

        private:
            static NOvACheckSum* _instanceID; //!< Instance pointer for the singleton class

            nova_crc_32_type _CRC32;          //!< CRC-32C object. Used when producing CRC
    };// end of class NOvACheckSum


}// end of daqdataformats namespace
#endif /* NOVA_CHECKSUM_H */
