#ifndef RAWCONFIGURATIONHEADER_H
#define RAWCONFIGURATIONHEADER_H
#include "DAQDataFormats/RawConfigurationHeaderV1.h"

/// It is requred to
///   1. include DAQDataFormats/DataFormatFirst.inc
///   2. define CURRENT_CLASS
///   3.    define CURRENT_CLASS_VERSION for version of the class
///      OR define LAST_CLASS_VERSION    for the main structure of the class
///   4. (optional) define CURRENT_FIELDS_SUB_NAMESPACE
///   5. include DAQDataFormats/DataFormatLast.inc
/// before the declaration of the DataFormat version class in the header.
#include "DAQDataFormats/DataFormatFirst.inc"
#define  CURRENT_CLASS                RawConfigurationHeader
#define  LAST_CLASS_VERSION           1
#define  CURRENT_FIELDS_SUB_NAMESPACE rawconfhdr
#include "DAQDataFormats/DataFormatLast.inc"

namespace daqdataformats {

    namespace rawconfhdr {
        static const version_t RAWCONFIGURATIONBLOCKHEADER_LATEST_VERSION = LAST_CLASS_VERSION;
    }

    class RawConfigurationHeader: public PREVIOUS_VERSION_CLASS_NAME{
        public:

            /// Constructors
            RawConfigurationHeader (const version_t version = rawconfhdr::RAWCONFIGURATIONBLOCKHEADER_LATEST_VERSION);                             //! The default constructor
            RawConfigurationHeader (const RawConfigurationHeader& copy_in);  //! Deep copy constructor
            RawConfigurationHeader (const void* dadablockheader_buffer, const version_t version = rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION); //! Shallow copy constructor
            /// Destructor
            ~RawConfigurationHeader();

            IMPLEMENT_MAIN_STANDARD
                    
            IMPLEMENT_GET_MAIN        (ConfigurationHeaderVersion, uint32_t);
            IMPLEMENT_GET_MAIN        (MarkerHi1            , uint32_t);
            IMPLEMENT_GET_MAIN        (MarkerHi2            , uint32_t);

            IMPLEMENT_GET_AND_SET_MAIN(DetectorID            , uint32_t);
            IMPLEMENT_GET_AND_SET_MAIN(ConfigurationBlockSize, uint32_t);
            IMPLEMENT_GET_AND_SET_MAIN(ResourceManagerID     , uint32_t)
            IMPLEMENT_GET_AND_SET_MAIN(RunNumber             , uint32_t)
            IMPLEMENT_GET_AND_SET_MAIN(RunType               , uint32_t)
            IMPLEMENT_GET_AND_SET_MAIN(ConfigurationIDX      , uint32_t)
            IMPLEMENT_GET_AND_SET_MAIN(Partition             , uint32_t)
            IMPLEMENT_GET_AND_SET_MAIN(Subrun                , uint32_t)
            IMPLEMENT_GET_AND_SET_MAIN(SystemNumber          , uint32_t)
            IMPLEMENT_GET_AND_SET_MAIN(SystemVersionIDX      , uint32_t)

            IMPLEMENT_FIGUREOUT_VERSION_STATIC_WITH_CUSTOM_GET_VERSION(RawConfigurationHeader, getConfigurationHeaderVersion)

    private:
            IMPLEMENT_SET_MAIN        (ConfigurationHeaderVersion, uint32_t);
            IMPLEMENT_SET_MAIN        (MarkerHi1            , uint32_t);
            IMPLEMENT_SET_MAIN        (MarkerHi2            , uint32_t);

            bool setFunctionPointers();



    };
  
} // END NAMESPCE daqdataformats
#endif /* RAWCONFIGURATIONHEADER_H */
