#ifndef RAWCONFIGURATIONHEADERV0_H
#define RAWCONFIGURATIONHEADERV0_H
#include "DAQDataFormats/RawDAQData.h"
#include "DAQDataFormats/Macros.h"
#include "DAQDataFormats/BitFields.h"

/// It is requred to
///   1. include DAQDataFormats/DataFormatFirst.inc
///   2. define CURRENT_CLASS
///   3.    define CURRENT_CLASS_VERSION for version of the class
///      OR define LAST_CLASS_VERSION    for the main structure of the class
///   4. (optional) define CURRENT_FIELDS_SUB_NAMESPACE
///   5. include DAQDataFormats/DataFormatLast.inc
/// before the declaration of the DataFormat version class in the header.
#include "DAQDataFormats/DataFormatFirst.inc"
#define  CURRENT_CLASS                RawConfigurationHeader
#define  CURRENT_CLASS_VERSION        0
#define  CURRENT_FIELDS_SUB_NAMESPACE rawconfhdr
#include "DAQDataFormats/DataFormatLast.inc"

namespace daqdataformats {

    namespace CURRENT_FIELDS_SUB_NAMESPACE {
        static const uint32_t Marker_value[2] = {0x464E4F43, 0x54525453};
    }

namespace  VERSION_NAMESPACE {
    namespace CURRENT_FIELDS_SUB_NAMESPACE {


            static const uint32_t RAWCONFIGURATIONBLOCKHEADER_SIZE = 7;
            
            enum RunHeaderMASKS{
                MASK_MarkerHi1                  = 0xFFFFFFFF, //! Marker Mask upper word 1      (32 bits)
                MASK_MarkerHi2                  = 0xFFFFFFFF, //! Marker Mask upper word 2      (32 bits)
                MASK_ConfigurationHeaderVersion = 0x00000FFF, //! Configuration header version  (12 bits)
                MASK_DetectorID                 = 0x00007000, //! Detector ID                   (3 bits)
                MASK_ConfigurationBlockSize     = 0xFFFF0000, //! Size of the Configuration Block in words (16 bits)
                MASK_ResourceManagerID          = 0x0000FFFF, //! Recource manager identifier   (16 bits)
                MASK_RunNumber                  = 0x00FFFFFF, //! Run number               (24 bits)
                MASK_RunType                    = 0xFF000000, //! Run type                 (8 bits)
                MASK_ConfigurationIDX           = 0x0000FFFF, //! Configuration IDX (Configuration Database table number) (16 bits)
                MASK_Partition                  = 0x00FF0000, //! Partition identifier that Run Control and Data Logger belong to (8 bits)
                MASK_Subrun                     = 0x3F000000, //! Subrun number            (6 bits)
                MASK_SystemNumber               = 0x0000FFFF, //! Number of systems        (16 bits)
                MASK_SystemVersionIDX           = 0xFFFF0000, //! System versionindex key  (16 bits)
            };

            enum RunHeaderWORDS{
                WORD_MarkerHi1                  = 0, //! Marker Mask upper word (32 bits)
                WORD_MarkerHi2                  = 1, //! Marker Mask lower word (32 bits)
                WORD_ConfigurationHeaderVersion = 2, //! Configuration header version  (12 bits)
                WORD_DetectorID                 = 2,
                WORD_ConfigurationBlockSize     = 2, //! Size of the Configuration Block in words (16 bits)
                WORD_ResourceManagerID          = 3, //! Recource manager identifier   (16 bits)
                WORD_RunNumber                  = 4, //! Run number             (24 bits)
                WORD_RunType                    = 4, //! Run type               (8 bits)
                WORD_ConfigurationIDX           = 5, //! Configuration IDX (Configuration Database table number) (16 bits)
                WORD_Partition                  = 5, //! Partition identifier that Run Control and Data Logger belong to (8 bits)
                WORD_Subrun                     = 5, //! Subrun number          (6 bits)
                WORD_SystemNumber               = 6, //! Number of systems
                WORD_SystemVersionIDX           = 6, //! System versionindex key
            };

            DEFINE_SHIFT_FROM_MASK(MarkerHi1);
            DEFINE_SHIFT_FROM_MASK(MarkerHi2);
            DEFINE_SHIFT_FROM_MASK(ConfigurationHeaderVersion);
            DEFINE_SHIFT_FROM_MASK(DetectorID);
            DEFINE_SHIFT_FROM_MASK(ConfigurationBlockSize);
            DEFINE_SHIFT_FROM_MASK(ResourceManagerID);
            DEFINE_SHIFT_FROM_MASK(RunNumber);
            DEFINE_SHIFT_FROM_MASK(RunType);
            DEFINE_SHIFT_FROM_MASK(ConfigurationIDX);
            DEFINE_SHIFT_FROM_MASK(Partition);
            DEFINE_SHIFT_FROM_MASK(Subrun);
            DEFINE_SHIFT_FROM_MASK(SystemNumber);
            DEFINE_SHIFT_FROM_MASK(SystemVersionIDX);
    }// end of namespace rawconfhdr


  class RawConfigurationHeader : public PREVIOUS_VERSION_CLASS_NAME {
    
        public:
            RawConfigurationHeader (const version_t); ///< The default constructor
            RawConfigurationHeader (const RawConfigurationHeader&);
            ~RawConfigurationHeader(); ///< Destructor

            void     init            ();
            using    PREVIOUS_VERSION_CLASS_NAME::print;
            void     print           (std::ostream& os=std::cout) const;
            uint32_t sizeofdata      () const {return CURRENT_FIELDS_SUB_NAMESPACE::RAWCONFIGURATIONBLOCKHEADER_SIZE;}
            using    PREVIOUS_VERSION_CLASS_NAME::readData;
            void*    readData(const void* pointer){return RawDAQData::readDataGeneral(pointer);};
            
            IMPLEMENT_FIRST_GET_AND_SET(ConfigurationHeaderVersion, uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(MarkerHi1, uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(MarkerHi2, uint32_t)

            /// Both get and set methods are public
            IMPLEMENT_FIRST_GET_AND_SET(DetectorID            , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(ConfigurationBlockSize, uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(ResourceManagerID     , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(RunNumber             , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(RunType               , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(ConfigurationIDX      , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(Partition             , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(Subrun                , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(SystemNumber          , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(SystemVersionIDX      , uint32_t)

	    INIT_FIRST(advanceSystemNumber, , bool, )

  };
  
  
 }} // END NAMESPCE daqdataformats
#endif /* RAWCONFIGURATIONHEADERV0_H */
