#ifndef RAWEVENTHEADERV0_H
#define RAWEVENTHEADERV0_H
#include "DAQDataFormats/RawDAQData.h"
#include "DAQDataFormats/Macros.h"
#include "DAQDataFormats/BitFields.h"

/// It is requred to
///   1. include DAQDataFormats/DataFormatFirst.inc
///   2. define CURRENT_CLASS
///   3.    define CURRENT_CLASS_VERSION for version of the class
///      OR define LAST_CLASS_VERSION    for the main structure of the class
///   4. (optional) define CURRENT_FIELDS_SUB_NAMESPACE
///   5. include DAQDataFormats/DataFormatLast.inc
/// before the declaration of the DataFormat version class in the header.
#include "DAQDataFormats/DataFormatFirst.inc"
#define  CURRENT_CLASS                RawEventHeader
#define  CURRENT_CLASS_VERSION        0
#define  CURRENT_FIELDS_SUB_NAMESPACE raweventhdr
#include "DAQDataFormats/DataFormatLast.inc"

namespace daqdataformats {

    namespace CURRENT_FIELDS_SUB_NAMESPACE {
        /// Event header delimiter
        static const uint32_t Marker_value[2] = {0xE929, 0xE929};
    }

namespace  VERSION_NAMESPACE {
   namespace CURRENT_FIELDS_SUB_NAMESPACE {

        const version_t VERSION_Tail = 0;

        const uint32_t RAWEVENTHEADER_SIZE = 8;

        enum EventHeaderMASKS{
            MASK_MarkerRight         = 0x0000FFFF, //! Marker Mask left                  (16 bits)
            MASK_MarkerLeft          = 0xFFFF0000, //! Marker Mask right                 (16 bits)
            MASK_EventSize           = 0xFFFFFFFF, //! Size of the event in words        (8 bits)
            MASK_Version             = 0x00000FFF, //! Version Mask                      (12 bits)
            MASK_DataLoggerID        = 0xFFFF0000, //! DataLogger ID                     (16 bits)
            MASK_RunNumber           = 0x00FFFFFF, //! Run number                        (24 bits)
            MASK_RunType             = 0xFF000000, //! Run type                          (8 bits)
            MASK_ConfigIDX           = 0x0000FFFF, //! Detector configuration identifier (16 bits)
            MASK_Partition           = 0x00FF0000, //! Data logger partition             (8 bits)
            MASK_Subrun              = 0x3F000000, //! SubRun number                     (6 bits)
            MASK_EventNumberLo       = 0xFFFFFFFF, //! Universal event number lower  0-31 bits
            MASK_EventNumberHi       = 0x000000FF, //! Universal event number upper 32-39 bits
            MASK_EventType           = 0xFF000000, //! Size of the event in words        (8 bits)
            MASK_DataBlockNumber     = 0x0000FFFF, //! Number of event size              (16 bits)
            MASK_EventIncomplete     = 0x20000000, //! Event incomplete flag             (1 bit)
            MASK_MonteCarloFlag      = 0x40000000, //! SIM flag                          (1 bit)
            MASK_CRCCalculationUsed  = 0x80000000  //! Is CRC calculation used?          (1 bit)
        };

        enum EventHeaderWORDS{
            WORD_MarkerRight         = 0,
            WORD_MarkerLeft          = 0,
            WORD_EventSize           = 1,
            WORD_Version             = 2,
            WORD_DataLoggerID        = 2,
            WORD_RunNumber           = 3,
            WORD_RunType             = 3,
            WORD_ConfigIDX           = 4,
            WORD_Partition           = 4,
            WORD_Subrun              = 4,
            WORD_EventNumberLo       = 5,
            WORD_EventNumberHi       = 6,
            WORD_EventType           = 6,
            WORD_DataBlockNumber     = 7,
            WORD_EventIncomplete     = 7, //! Event incomplete flag             (1 bit)
            WORD_MonteCarloFlag      = 7,
            WORD_CRCCalculationUsed  = 7
        };

            DEFINE_SHIFT_FROM_MASK(MarkerRight         )
            DEFINE_SHIFT_FROM_MASK(MarkerLeft          )
            DEFINE_SHIFT_FROM_MASK(EventSize           )
            DEFINE_SHIFT_FROM_MASK(Version             )
            DEFINE_SHIFT_FROM_MASK(DataLoggerID        )
            DEFINE_SHIFT_FROM_MASK(RunNumber           )
            DEFINE_SHIFT_FROM_MASK(RunType             )
            DEFINE_SHIFT_FROM_MASK(ConfigIDX           )
            DEFINE_SHIFT_FROM_MASK(Partition           )
            DEFINE_SHIFT_FROM_MASK(Subrun              )
            DEFINE_SHIFT_FROM_MASK(EventNumberLo       )
            DEFINE_SHIFT_FROM_MASK(EventNumberHi       )
            DEFINE_SHIFT_FROM_MASK(EventType           )
            DEFINE_SHIFT_FROM_MASK(DataBlockNumber     )
            DEFINE_SHIFT_FROM_MASK(EventIncomplete     )
            DEFINE_SHIFT_FROM_MASK(MonteCarloFlag      )
            DEFINE_SHIFT_FROM_MASK(CRCCalculationUsed  )


  }// end of namespace raweventhdr
  
    class RawEventHeader: public PREVIOUS_VERSION_CLASS_NAME {
        public:
            RawEventHeader (const version_t); ///< The default constructor
            RawEventHeader (const RawEventHeader&);
            ~RawEventHeader(); ///< Destructor

            void     init            ();
            using    PREVIOUS_VERSION_CLASS_NAME::print;
            void     print           (std::ostream& os=std::cout) const;
            uint32_t sizeofdata      () const { return CURRENT_FIELDS_SUB_NAMESPACE::RAWEVENTHEADER_SIZE;}
            using    RawDAQData::readData;
            void*    readData(const void* pointer){return RawDAQData::readDataGeneral(pointer);};

            // Fields that are defined in later versions, but are not allowed in this version
            bool     setTriggerFilled(const uint32_t) {return GENERATE_FUNCTION_IS_NOT_ALLOWED;}
            bool     setTailFilled   (const uint32_t) {return GENERATE_FUNCTION_IS_NOT_ALLOWED;}
            uint32_t getTriggerFilled() const         {GENERATE_FUNCTION_IS_NOT_ALLOWED; return 0;}
            uint32_t getTailFilled   () const         {GENERATE_FUNCTION_IS_NOT_ALLOWED; return 0;}

        protected:

            IMPLEMENT_FIRST_GET_AND_SET(MarkerLeft      , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(MarkerRight     , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(Version         , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(DataLoggerID    , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(RunNumber       , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(RunType         , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(ConfigIDX       , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(Partition       , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(Subrun          , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(EventNumberLo   , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(EventNumberHi   , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(EventType       , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(EventSize       , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET(DataBlockNumber , uint32_t)
            IMPLEMENT_FIRST_GET_AND_SET_BOOL(MonteCarloFlag)
            IMPLEMENT_FIRST_IS_AND_SET      (CRCCalculationUsed)
            IMPLEMENT_FIRST_IS_AND_SET      (EventIncomplete)

            INIT_FIRST(addEventSize          , ,bool       , const uint32_t)
	    INIT_FIRST(advanceEventNumber    , ,bool, )
	    INIT_FIRST(advanceDataBlockNumber, ,bool, )
            
	    INIT_FIRST(getEventNumber        , const, uint64_t, )
            INIT_FIRST(setEventNumber        ,      , bool       , const uint64_t)

            /// Returns tail version corresponding to the Event(EventHeader) version
	    INIT_FIRST(getTailVersion        , const, version_t, )

    };
  
  
}} // END NAMESPCE daqdataformats
#endif /* RAWEVENTHEADERV0_H */
