#ifndef _DBUTILS_SIMPLEPARAMETERSET_H
#define _DBUTILS_SIMPLEPARAMETERSET_H

#include <boost/lexical_cast.hpp>
#include <map>
#include <string>
#include <vector>

namespace dbutils {

/**
 * Simple class to set and retrieve key/value pairs.
 *
 * This class is somewhat database-specific in its handling of
 * true/false values.
 *
 * @author Kurt Biery
 * @version $Revision: 1.5 $ $Date: 2011/10/31 19:38:09 $
 */

class SimpleParameterSet {

 public:

    // *****************
    // *** Modifiers ***
    // *****************

    void put(const std::string& key, const std::string& value)
    {
        _paramSet[key] = value;
    }

    void clear()
    {
        _paramSet.clear();
    }

    // *****************
    // *** Accessors ***
    // *****************

    template <class T>
    bool get(const std::string& key, T& resultValue) const
    {
        std::map<std::string, std::string>::const_iterator iter =
            _paramSet.find(key);
        if (iter == _paramSet.end()) {return false;}

        try {
            resultValue = boost::lexical_cast<T>(iter->second);
            return true;
        }
        catch (...) {
            // fall through to "false" (failing) return value below
        }
        return false;
    }

    bool get(const std::string& key, bool& resultValue) const
    {
        std::map<std::string, std::string>::const_iterator iter =
            _paramSet.find(key);
        if (iter == _paramSet.end()) {return false;}

        if (iter->second == "t" || iter->second == "1") {
            resultValue = true;
        }
        else {
            resultValue = false;
        }
        return true;
    }

    std::vector<std::string> keys() const;

    std::string toString() const;

 private:

    std::map<std::string, std::string> _paramSet;
};

};  // end of namespace dbutils

#endif
