#include "DAQDataFormats/RawMicroBlockHeaderV1.h"
#include <boost/format.hpp>

/// It is requred to
///   1. include DAQDataFormats/DataFormatFirst.inc
///   2. define CURRENT_CLASS
///   3.    define CURRENT_CLASS_VERSION for version of the class
///      OR define LAST_CLASS_VERSION    for the main structure of the class
///   4. (optional) define CURRENT_FIELDS_SUB_NAMESPACE
///   5. include DAQDataFormats/DataFormatLast.inc
/// before the declaration of the DataFormat version class in the header.
#include "DAQDataFormats/DataFormatFirst.inc"
#define  CURRENT_CLASS                RawMicroBlockHeader
#define  CURRENT_CLASS_VERSION        1
#define  CURRENT_FIELDS_SUB_NAMESPACE rawmicroblockhdr
#include "DAQDataFormats/DataFormatLast.inc"

namespace daqdataformats{
namespace VERSION_NAMESPACE{

//------------------------------------------------------------
// Default constructor
RawMicroBlockHeader::RawMicroBlockHeader(const version_t version):
     PREVIOUS_VERSION_CLASS_NAME(version)
    ,ZEROOUT_GET_AND_SET(V1)
{ }

//------------------------------------------------------------
RawMicroBlockHeader::RawMicroBlockHeader(const RawMicroBlockHeader& copyin):
     PREVIOUS_VERSION_CLASS_NAME(copyin)
    ,DEEPCOPY_GET_AND_SET(V1)
{ }

//------------------------------------------------------------
// Destructor
RawMicroBlockHeader::~RawMicroBlockHeader() { }

//------------------------------------------------------------
uint32_t RawMicroBlockHeader::getVersion()  const {
        const uint32_t v0 = EXECUTE_FUNCTION(getV0)();
        const uint32_t v1 = EXECUTE_FUNCTION(getV1)();
        return v0 + v1;
}

//------------------------------------------------------------
bool     RawMicroBlockHeader::setVersion(const uint32_t version){
    /// At this point, version cannot be zero
    EXECUTE_FUNCTION(setV0)(1);
    RETURN_FUNCTION(setV1)(version - 1);
}


//------------------------------------------------------------------------------
// Print the MicroBlock header information
void RawMicroBlockHeader::print(std::ostream& os) const{
    os<<"\nMicroBlock Header:";
    os<<boost::format(" getDiblockId=%u")% getDiblockId();
    os<<boost::format(" DCMId=%u"       )% getDCMId();
    os<<boost::format(" DetId=%u"       )% getDetId();
    os<<boost::format(" DestBuff=%u"    )% EXECUTE_FUNCTION(getDestBuffID)();   //! Destination buffer node
    os<<boost::format(" Partition=%u"   )% EXECUTE_FUNCTION(getPartition)();    //! Partition
    os<<boost::format(" V0=%u"          )% EXECUTE_FUNCTION(getV0)();
    os<<boost::format(" V1=%u"          )% EXECUTE_FUNCTION(getV1)();
    os<<boost::format(" Version=%u"     )% EXECUTE_FUNCTION(getVersion)();
    os<<"\n";

    printBuffer(os);

    const uint32_t size_of_data = EXECUTE_FUNCTION(sizeofdata)();

    os<<"\nMicroBlockHeader RAW 32bit words (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < size_of_data; ++i){
        os<<" ";
        printWord(i, false, os);
    }

    os<<"\nMicroBlockHeader RAW bits (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < size_of_data; ++i){
        os<<std::endl;
        printWord(i, true, os);
    }

}


}}// end of namespaces