#include "DAQDataFormats/RawNanoSliceV2.h"
#include <boost/format.hpp>

/// It is requred to
///   1. include DAQDataFormats/DataFormatFirst.inc
///   2. define CURRENT_CLASS
///   3.    define CURRENT_CLASS_VERSION for version of the class
///      OR define LAST_CLASS_VERSION    for the main structure of the class
///   4. (optional) define CURRENT_FIELDS_SUB_NAMESPACE
///   5. include DAQDataFormats/DataFormatLast.inc
/// before the declaration of the DataFormat version class in the header.
#include "DAQDataFormats/DataFormatFirst.inc"
#define  CURRENT_CLASS                RawNanoSlice
#define  CURRENT_CLASS_VERSION        2
#define  CURRENT_FIELDS_SUB_NAMESPACE rawnano
#include "DAQDataFormats/DataFormatLast.inc"
  
namespace daqdataformats{
namespace VERSION_NAMESPACE{
//------------------------Constructors------------------------------------------
RawNanoSlice::RawNanoSlice(const version_t version):
             PREVIOUS_VERSION_CLASS_NAME(version)
            ,ZEROOUT_GET_AND_SET(Value)
{ }
RawNanoSlice::RawNanoSlice(const NanoSlice& nano):
             PREVIOUS_VERSION_CLASS_NAME(nano)
            ,ZEROOUT_GET_AND_SET(Value)
{ }
//------------------------------------------------------------------------------
RawNanoSlice::RawNanoSlice (const RawNanoSlice& copyin):
             PREVIOUS_VERSION_CLASS_NAME(copyin)
            ,DEEPCOPY_GET_AND_SET(Value)
{}

//------------------------Destructor--------------------------------------------
RawNanoSlice::~RawNanoSlice() { }


//------------------------------------------------------------------------------
// Default Initialization of Data members
void RawNanoSlice::init(){
    /// After using resetBufferSource header is already constructed
    /// Now we need to initialize it
    _Header->init();
}

//------------------------------------------------------------------------------
void RawNanoSlice::initWithData(const NanoSlice& nanoslice_parameters){

    /// After using resetBufferSource header is already constructed
    /// Now we need to initialize it
    _Header->initWithData(nanoslice_parameters.Header);

    EXECUTE_FUNCTION(setTimeStamp)(nanoslice_parameters.TimeStamp);
    //EXECUTE_FUNCTION(setValue1)(nanoslice_parameters.Value1);
    //EXECUTE_FUNCTION(setValue2)(nanoslice_parameters.Value2);
    
}// end of initWithData

//-----------------------------------------------------------------
// Set all parameters
bool RawNanoSlice::setParameters(const NanoSlice& nanoslice_parameters){
  
        _Header->setParameters(nanoslice_parameters.Header);

        EXECUTE_FUNCTION(setTimeStamp)(nanoslice_parameters.TimeStamp);
        //EXECUTE_FUNCTION(setValue1)(nanoslice_parameters.Value1);
        //EXECUTE_FUNCTION(setValue2)(nanoslice_parameters.Value2);

        return true;
}

//------------------------------------------------------------------------------
uint32_t RawNanoSlice::getValue(const uint32_t sample_number) const
{
    uint32_t mask, shift;
    const uint32_t* pointer = getWordPointer(sample_number, mask, shift);

    return BITFIELDGET(*pointer, mask, shift);
}


//------------------------------------------------------------------------------
bool RawNanoSlice::setValue(const uint32_t sample_number, const uint32_t value)
{
    uint32_t mask, shift;
    uint32_t* pointer = getWordPointer(sample_number, mask, shift);

    BITFIELDSET(*pointer, value, mask, shift);
    return true;
}

//------------------------------------------------------------------------------
// Returns pointer to the word, also mask and shift
uint32_t* RawNanoSlice::getWordPointer(const uint32_t sample_number, uint32_t& mask, uint32_t& shift) const
{

    const uint32_t  sample_number_in_word = sample_number % 2;

    if(sample_number_in_word == 1){
        mask  = version2::rawnano::MASK_Value2;
        shift = version2::rawnano::SHIFT_Value2;
    }
    else{
        mask  = version1::rawnano::MASK_Value1;
        shift = version1::rawnano::SHIFT_Value1;
    }

    const uint32_t  word_number           = version0::rawnano::WORD_Reserved
                                          + sample_number / 2;

    return  &(RAWBUFF32[word_number]);
}

//------------------------------------------------------------------------------
// Printing the NanoSlice
void RawNanoSlice::print(std::ostream& os) const
{

    if(_version == 0) return version0::RawNanoSlice::print(os);
    if(_version == 1) return version1::RawNanoSlice::print(os);

    const uint32_t size_of_data = EXECUTE_FUNCTION(sizeofdata)();

    os<<"NanoSlice Version "<<_version<<" has size "<<size_of_data<<"\n";

    os<<"\nRAW(hex):";
    for(uint32_t i = 0; i < size_of_data; ++i){
            printWord(i, false, os);
            os<<" ";
    }

    os<<"Printing Header\n";
    _Header->print(os);

    os<<"Getting timestamp\n";
    os<<boost::format(" NanoSlice Body: TimeStamp=%u")
	 %EXECUTE_FUNCTION(getTimeStamp)()
	 ;



    const int nsamples = getNSamples();
    os<<"\n Total of "<<nsamples<<" samples:";
    for(int i=0; i<nsamples; ++i){
            os<<" Sample "<<i<<":"<<EXECUTE_FUNCTION(getValue)(i);
    }



    os<<"\nRAW(binary):";
    for(uint32_t i = 0; i < size_of_data; ++i){
            if(i > 0)os<<"    ";
            printWord(i, true, os);
            os<<std::endl;
    }



}// end of print

}}// end of namespaces
