#include <DatabaseUtils/DAQAppMgr/ConfiguredHosts.h>
#include <NovaDatabase/Table.h>
#include <NovaDatabase/Util.h>
#include <boost/shared_ptr.hpp>

namespace NOVADB = nova::database;

namespace dbutils {
namespace daqappmgr {

/**
 * Creates a new instance based on the specified detector name
 * (NDOS, NearDet, FarDet) and partition number
 */
ConfiguredHosts::
ConfiguredHosts(const std::string& detectorName, int partitionNumber,
                const bool& useDBData)
{
    _hasValidData = false;
    std::string fname;
}

/**
 * Fetches the host with the specified name.
 */
boost::shared_ptr<Host> ConfiguredHosts::
getHostByName(const std::string& hostName) const
{
    boost::shared_ptr<Host> hostPtr;
    if (! _hasValidData) {return hostPtr;}

    for (int idx = 0; idx < (int) _hostList.size(); ++idx) {
        if (_hostList[idx]->getHostName() == hostName) {
            hostPtr = _hostList[idx];
            break;
        }
    }

    return hostPtr;
}

/**
 * Adds the specified type to the list
 */
void ConfiguredHosts::
_addHost(boost::shared_ptr<Host> host)
{
    if (host.get() == 0) {return;}

    _hasValidData = true;

    _hostList.push_back(host);
}

};  // end of namespace daqappmgr
};  // end of namespace dbutils
