#include <DatabaseUtils/DAQAppMgr/Host.h>

namespace dbutils {
namespace daqappmgr {

const std::string Host::MANAGER_NODE_TYPE("manager");
const std::string Host::FARM_NODE_TYPE("farm");
const std::string Host::REAL_DCM_TYPE("dcm");

/**
 * Creates an instance from the specified values.
 */
Host::Host(const std::string& hostName, const std::string& hostType,
           const std::string& sshUserName, const bool& ppcFlag,
           const bool& noKerberosFlag) :
    _hostName(hostName), _hostType(hostType), _sshUserName(sshUserName),
    _ppcFlag(ppcFlag), _noKerberosFlag(noKerberosFlag)
{
}

/**
 * Tests if the specified instance is equal to this one.
 */
bool Host::
operator==(const Host& other) const
{
    if (other._ppcFlag && ! this->_ppcFlag) {return false;}
    if (! other._ppcFlag && this->_ppcFlag) {return false;}

    if (other._noKerberosFlag && ! this->_noKerberosFlag) {return false;}
    if (! other._noKerberosFlag && this->_noKerberosFlag) {return false;}

    return (other._hostName == this->_hostName &&
            other._hostType == this->_hostType &&
            other._sshUserName == this->_sshUserName);
}

/**
 * Tests if the specified set is not equal to this one.
 */
bool Host::
operator!=(const Host& other) const
{
    return !(*this==other);
}

/**
 * Tests if the specified instance is less than this one.
 */
bool Host::
operator<(const Host& other) const
{
    if (this->_hostName > other._hostName) {return false;}
    if (this->_hostName < other._hostName) {return true;}
    // else _hostName is equal

    if (this->_hostType > other._hostType) {return false;}
    if (this->_hostType < other._hostType) {return true;}
    // else _hostType is equal

    if (this->_sshUserName > other._sshUserName) {return false;}
    if (this->_sshUserName < other._sshUserName) {return true;}
    // else _sshUserName is equal

    if (! this->_ppcFlag && other._ppcFlag) {return true;}
    if (this->_ppcFlag && ! other._ppcFlag) {return false;}

    if (! this->_noKerberosFlag && other._noKerberosFlag) {return true;}
    else {return false;}
}

};  // end of namespace daqappmgr
};  // end of namespace dbutils
