#include <NovaDAQUtilities/Cache.h>

namespace novadaq {

/**
 * Flushes the cache.  (Removes any saved data.)
 */
void Cache::flush() {
    boost::mutex::scoped_lock lock(_cacheMutex);
    _clearCachedData();
    _cacheIsEmpty = true;
}

/**
 * Constructor.
 *
 * @param policy the CachePolicy to use for automatically expiring
 *        the cache.
 */
Cache::Cache(CachePolicy policy) : _policy(policy), _cacheIsEmpty(true) {}

/**
 * Reloads the cache if needed.
 */
void Cache::_reloadIfNeeded() {
    if (_cacheIsEmpty || _policy.isExpired()) {
        _clearCachedData();
        _loadDataIntoCache();
        _cacheIsEmpty = false;
        _policy.cacheUpdated();
    }
}

};  // end of namespace novadaq
